import os
import copy
import numpy as np
import dopyqo
import warnings
import logging
import pandas as pd
import dopyqo
from dopyqo.colors import *
from qiskit.quantum_info import Statevector
from dopyqo import cube


if __name__ == "__main__":
    # Run QE SCF calculation
    # and run QE pp calculation
    # using the input files Mg2Si.scf.in and Mg2Si.pp.in

    base_folder = os.path.join("qe_files", "Mg2Si")
    prefix = "Mg2Si"
    active_electrons = 8
    active_orbitals = 9
    path_to_bader_executable = {"bader": os.path.join(os.path.expanduser("~"), "bader/bader")}
    config = dopyqo.DopyqoConfig(
        base_folder=base_folder,
        prefix=prefix,
        active_electrons=active_electrons,
        active_orbitals=active_orbitals,
        run_fci=True,
        run_vqe=False,
        vqe_optimizer=dopyqo.VQEOptimizers.L_BFGS_B,
        n_threads=10,
        uccsd_reps=1,
    )

    energy_dict, wfc1_ncpp, h_ks, mats = dopyqo.run(config)

    npz_path = os.path.join(f"{base_folder}", f"statevecs_{active_electrons}e_{active_orbitals}o.npz")
    atoms = []
    atoms = [element["element"] for element in wfc1_ncpp.atoms]
    valence = {x.element: x.z_valence for x in wfc1_ncpp.pseudopots}
    # Extracting data from the ACF.dat file in the pp-x_bader folder. WARNING here pp is related to the pp.x command no PseudoPotential

    reference_energy = wfc1_ncpp.etot

    overlaps_ncpp = wfc1_ncpp.get_overlaps()

    if not np.allclose(overlaps_ncpp, np.eye(overlaps_ncpp.shape[0])):
        warnings.warn("The wavefunctions are not orthonormal!")

    p = wfc1_ncpp.k_plus_G  # shape (#waves, 3)
    c_ip = wfc1_ncpp.evc  # shape (#bands, #waves)

    logging.info("Number of plane waves: %i", p.shape[0])

    ########################################################
    #           Calculating real-space densities           #
    ########################################################
    nbands = wfc1_ncpp.nbnd
    max_min = wfc1_ncpp.fft_grid
    mill = wfc1_ncpp.mill
    tot_elec = wfc1_ncpp.nelec

    c_ip_array = np.zeros(
        (
            nbands,
            *max_min,
        ),
        dtype=c_ip.dtype,
    )

    # Set \psi_i(p) on given grid points
    assert mill.shape[0] == c_ip.shape[1]
    for idx, mill_idx in enumerate(mill):
        x, y, z = mill_idx
        i, j, k = (
            x + max_min[0] // 2,
            y + max_min[1] // 2,
            z + max_min[2] // 2,
        )
        c_ip_array[:, i, j, k] = c_ip[:, idx]

    psi_r = []
    WF_cube = []
    QE_file = cube.read_cube(os.path.join(base_folder, f"{prefix}_pp.cube"))
    QE_cube = QE_file["cube"]
    QE_vol = QE_file["volume"]
    QE_cube_scale = QE_cube * QE_vol

    print(f"nbands: {nbands}")
    for i in range(nbands):
        logging.info("Calculating real space KS %s/%s ...", i + 1, nbands)
        c_ip_shifted = np.fft.ifftshift(c_ip_array[i])
        psi_r_i = np.fft.ifftn(c_ip_shifted) * np.sqrt(np.prod(max_min))

        psi_r.append(psi_r_i)

    psi_r = np.array(psi_r)

    occ_bands = int(tot_elec / 2)
    ks_occupation = []

    for i in range(occ_bands):
        occ = 1
        ks_occupation.append(occ)

    for i in range(nbands - occ_bands):
        occ = 0
        ks_occupation.append(occ)

    ks_single_density = np.abs(psi_r) ** 2

    ks_tot_density = cube.get_density_cube(ks_single_density, wfc1_ncpp.occupations_xml)
    ks_tot_density = ks_tot_density / (wfc1_ncpp.cell_volume / np.prod(wfc1_ncpp.fft_grid) / 2.0)
    ks_total_electrons = wfc1_ncpp.nelec

    save_dir = os.path.join(base_folder, f"Bader_dopyqo")
    cube.make_cube(base_folder, prefix, ks_tot_density, save_dir, path_to_bader_executable, max_min)
    ks_ACF_file = os.path.join(save_dir, "ACF.dat")
    ks_bader = cube.extract_charge(atoms, ks_ACF_file, valence)

    ppx_dir = os.path.join(base_folder, "Bader_QE")
    ppx_file = os.path.join(base_folder, f"{prefix}_pp.cube")
    cube.run_ppx(ppx_file, path_to_bader_executable, ppx_dir, prefix)
    ppx_ACF_file = os.path.join(ppx_dir, "ACF.dat")
    qe_bader = cube.extract_charge(atoms, ppx_ACF_file, valence)

    # print(f"{ks_tot_density=}")
    print(f"{ks_total_electrons=}")

    QE_file = cube.read_cube(os.path.join(base_folder, f"{prefix}_pp.cube"))
    QE_cube = QE_file["cube"]
    QE_vol = QE_file["volume"]
    QE_cube_scale = QE_cube * QE_vol

    statevecs_data = np.load(npz_path)
    merged = pd.DataFrame()
    if statevecs_data["fci_state_data"].shape != (0,):
        fci_state_loaded = Statevector(statevecs_data["fci_state_data"])
        fci_AS_occupation = cube.orb_occupation(fci_state_loaded)
        fci_total_occupation = cube.make_full_occ_from_partial_occ(fci_AS_occupation, active_orbitals, active_electrons, ks_total_electrons, nbands)
        save_dir = os.path.join(base_folder, f"Bader_dopyqo_casci_{active_electrons}e_{active_orbitals}o")
        print(f"Occupation of AS of {prefix}: {fci_AS_occupation}")
        print(f"Occupation of CASCI of {prefix} : {fci_total_occupation}")
        print(f"Occupation KS of {prefix}: {wfc1_ncpp.occupations_xml}")
        fci_tot_density = cube.get_density_cube(ks_single_density, fci_total_occupation)
        fci_tot_density = fci_tot_density / (wfc1_ncpp.cell_volume / np.prod(wfc1_ncpp.fft_grid) / 2.0)
        cube.make_cube(base_folder, prefix, fci_tot_density, save_dir, path_to_bader_executable, max_min)
        fci_ACF_file = os.path.join(save_dir, "ACF.dat")
        fci_bader = cube.extract_charge(atoms, fci_ACF_file, valence)

        merged = pd.DataFrame(
            {
                "ELEMENT": atoms,
                "Bader charge QE": qe_bader["EFFECTIVE CHARGE"].values,
                "Bader charge Dopyqo": ks_bader["EFFECTIVE CHARGE"].values,
                "Bader charge Dopyqo CASCI": fci_bader["EFFECTIVE CHARGE"].values,
            }
        )
    if statevecs_data["vqe_state_data"].shape != (0,):
        vqe_state_loaded = Statevector(statevecs_data["vqe_state_data"])
        vqe_AS_occupation = cube.orb_occupation(vqe_state_loaded)
        vqe_total_occupation = cube.make_full_occ_from_partial_occ(vqe_AS_occupation, active_orbitals, active_electrons, ks_total_electrons, nbands)
        save_dir = os.path.join(base_folder, f"Bader_dopyqo_vqe/s_{active_electrons}e_{active_orbitals}o_")
        print(f"Occupation of AS of {prefix}: {vqe_AS_occupation}")
        print(f"Occupation of VQE of {prefix} : {vqe_total_occupation}")
        print(f"Occupation KS of {prefix}: {wfc1_ncpp.occupations_xml}")
        vqe_tot_density = cube.get_density_cube(ks_single_density, vqe_total_occupation)
        vqe_tot_density = vqe_tot_density / (wfc1_ncpp.cell_volume / np.prod(wfc1_ncpp.fft_grid) / 2.0)
        cube.make_cube(base_folder, prefix, vqe_tot_density, save_dir, path_to_bader_executable, max_min)
        vqe_ACF_file = os.path.join(save_dir, "ACF.dat")
        vqe_bader = cube.extract_charge(atoms, vqe_ACF_file, valence)

        merged = pd.DataFrame(
            {
                "ELEMENT": atoms,
                "Bader charge QE": qe_bader["EFFECTIVE CHARGE"].values,
                "Bader charge Dopyqo": ks_bader["EFFECTIVE CHARGE"].values,
                "Bader charge Dopyqo VQE": vqe_bader["EFFECTIVE CHARGE"].values,
            }
        )

    print(f"{GREEN}Bader charge results:{RESET_COLOR}")
    print(merged)



import sys
import logging
from warnings import warn
import os
import time
import copy
from enum import Enum
import numpy as np
from qiskit.quantum_info import Statevector
from qiskit_nature.second_q.problems.electronic_structure_result import (
    ElectronicStructureResult,
)
from qiskit_algorithms.minimum_eigensolvers.vqe import VQEResult
from qiskit.quantum_info import SparsePauliOp
from qiskit_nature.second_q.hamiltonians import ElectronicEnergy
from qiskit_nature.second_q.problems import ElectronicStructureProblem
from qiskit_nature.second_q.operators import ElectronicIntegrals
from qiskit_algorithms.minimum_eigensolvers import VQE, VQEResult
from qiskit_nature.second_q.mappers import JordanWignerMapper
from qiskit_nature.second_q.mappers.fermionic_mapper import FermionicMapper
import qiskit_algorithms.optimizers as optim
from qiskit.circuit import QuantumCircuit
from qiskit_nature.second_q.circuit.library import UCCSD, HartreeFock
from qiskit.primitives import Estimator, StatevectorEstimator
import pyscf.fci
from pyscf import gto, scf, cc
from pyscf.fci.cistring import str2addr
import tencirchem
import scipy
import dopyqo
from dopyqo import calc_matrix_elements
import dopyqo.fci_vector_matrix
from dopyqo.helpers import tcc_helpers
from dopyqo import fci_vector_matrix
from dopyqo.colors import *


def transform_index(i, n_orbitals):
    """From tequila order (up, down, up, down) to TenCirChem order (up, up, ..., down, down)"""
    i_orbital = i // 2
    i_spin = i % 2
    return i_orbital + n_orbitals * i_spin


class Hamiltonian:
    """Defines a second-quanitzation hamiltonian of the form

    H = \sum_{pq} h_{pq} a_p^\dagger a_q + 1/2 \sum_{pqrs} h_{pqrs} a_p^\dagger a_q^\dagger a_r a_s + E_{const.}

    with h_{pqrs} = \int \int \phi^*_p(r_1)  \phi^*_q(r_2) \phi_r(r_2)  \phi_s(r_1) / |r_1-r_2| dr_1 dr_2

    h_{pq} and h_{pqrs} are defined in terms of N spatial orbitals.

    Args:
        h_pq (np.ndarray): One-electron matrix elements. Shape (N, N).
        h_pqrs (np.ndarray): Two-electron matrix elements. Shape (N, N, N, N).
        occupations (np.ndarray): Occupations of the N spatial orbitals taking values either 0 or 1, and summing up to half of the number of electrons. Shape (N,)
        constants (float | dict[str, float], optional): Energy offset E_{const.} as float or dict of str, float pairs. Defaults to 0.0.
        reference_energy (float, optional): Reference energy used for VQE calculations. Defaults to 0.0.
    """

    def __init__(
        self,
        h_pq: np.ndarray,
        h_pqrs: np.ndarray,
        occupations: np.ndarray,
        constants: float | dict[str, float] = 0.0,
        reference_energy: float = 0.0,
    ):
        # TODO: Currently no support for spin-polarized h_pq, h_pqrs
        assert h_pq.ndim == 2
        assert h_pqrs.ndim == 4

        assert h_pq.shape[0] == h_pq.shape[1] == h_pqrs.shape[0] == h_pqrs.shape[1] == h_pqrs.shape[2] == h_pqrs.shape[3]

        # assert occupations.ndim == 1
        # assert occupations.shape[0] == h_pq.shape[0]

        self.norb = h_pq.shape[0]

        self.nspin = 1
        self.occupations = occupations
        for x in self.occupations:
            if not (np.isclose(int(x), 0) or np.isclose(int(x), 1)):
                print(
                    f"{RED}Hamiltonian error: occupations must be either 0 or 1 but found value {x}!{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
        self.nelec = round(np.sum(self.occupations))
        if self.nelec >= self.norb:
            print(
                f"{RED}Hamiltonian error: Half the number of electrons ({self.nelec}) is equal or greater, but has to be less, ",
                f"than the number of spatial orbitals ({self.norb})!{RESET_COLOR}",
            )
            sys.exit(1)

        logging.info("Checking matrix element symmetries...")
        if calc_matrix_elements.check_symmetry_one_body_matrix(h_pq) is False:
            print(f"{ORANGE}Hamiltonian warning: h_pq does not obey one body matrix symmetries (hermiticity)!{RESET_COLOR}")
        if False in (sym := calc_matrix_elements.check_symmetry_two_body_matrix(h_pqrs)):
            print(
                f"{ORANGE}Hamiltonian warning: h_pqrs does not obey two body matrix symmetries (swap symmetry {'fulfilled' if sym[0] else 'not fulfilled'}, "
                f"hermiticity {'fulfilled' if sym[1] else 'not fulfilled'}, hermiticity+swap {'fulfilled' if sym[2] else 'not fulfilled'})!{RESET_COLOR}"
            )

        self.h_pq = h_pq
        self.h_pqrs = h_pqrs
        if not isinstance(constants, dict) and not isinstance(constants, float):
            print(
                f"{RED}Hamiltonian error: Provided value for 'constants' is of type {type(constants)} but only dict or float are supported!{RESET_COLOR}"
            )
            sys.exit(1)
        if isinstance(constants, dict):
            key_types = set(map(type, constants.keys()))
            value_types = set(map(type, constants.values()))
            if key_types != {str} or value_types == {float}:
                print(
                    f"{RED}Hamiltonian error: Provided value for 'constants' is dict with keys of type {key_types} and values of type {value_types} "
                    + f"but only keys of type str and values of type float are supported!{RESET_COLOR}"
                )
                sys.exit(1)
        self.constants = constants
        self.reference_energy = reference_energy

        self.fci_energy = np.nan
        self.fci_solver: pyscf.fci.direct_spin1.FCIBase = pyscf.fci.direct_spin1.FCIBase()
        self.fci_evs: float = np.nan
        self.fci_evcs: list = []
        self.qiskit_elec_struc_result = ElectronicStructureResult()
        self.qiskit_vqe_result = VQEResult()
        self.qiskit_problem = None
        self.qiskit_ansatz = QuantumCircuit()
        self.qiskit_pauli_sum_op = SparsePauliOp(["I"])
        self.qiskit_vqe_counts = {}
        self.qiskit_vqe_values = {}
        self.qiskit_vqe_initial_state = None
        self.qiskit_vqe_ansatz = {}
        self.qiskit_vqe_solver = {}
        self.qiskit_vqe_result = None

        self.tcc_vqe_counts = None
        self.tcc_vqe_values = None
        self.tcc_vqe_result = None
        self.tcc_ansatz = None

        # Take the values of the latest ran VQE result, either from TenCirChem or qiskit
        self.vqe_counts = None
        self.vqe_values = None
        self.vqe_result = None

        # Used in solve_vqe when using TenCirChem and set in solve_hf
        self.mo_coeff = None
        self.mf = None

    def to_qiskit_problem(self, auto_index_order=True) -> ElectronicStructureProblem:
        """Create Qiskit ElectronicStructureProblem object.

        Args:
            auto_index_order (bool, optional): auto_index_order parameter passed to Qiskit ElectronicIntegrals.from_raw_integrals. Defaults to True.

        Returns:
            ElectronicStructureProblem: Created Qiskit ElectronicStructureProblem
        """
        num_particles = (
            self.nelec,
            self.nelec,
        )

        # Qiskit calculation
        integrals = ElectronicIntegrals.from_raw_integrals(self.h_pq, self.h_pqrs, auto_index_order=auto_index_order)
        qiskit_energy = ElectronicEnergy(integrals)

        if isinstance(self.constants, dict):
            for key, val in self.constants.items():
                qiskit_energy.constants[key] = val
        else:
            qiskit_energy.constants["energy_offset"] = self.constants
        # qiskit_energy.constants["frozen_core_energy"] = self.constants
        qiskit_problem = ElectronicStructureProblem(qiskit_energy)

        # number of particles for spin-up, spin-down
        qiskit_problem.num_particles = num_particles
        qiskit_problem.num_spatial_orbitals = self.norb

        qiskit_problem.reference_energy = self.reference_energy

        self.qiskit_problem = qiskit_problem

        return qiskit_problem

    def hf_energy(self) -> float:
        """Calculate the mean-field energy of the single Slater determinant given by self.occupations.
        If this Slater determinant is the Hartree-Fock state, the energy is the Hartree-Fock energy.

        The calculated energy is:
        E = 2 \sum_i h_{ii} + \sum_{ij) (2 h_{ijji} - h_{ijij}) + E_{const.}
        where the sums go over the occupied orbitals

        Returns:
            float: The calculated mean-field energy
        """
        h_pq = self.h_pq
        h_pqrs = self.h_pqrs

        energy_offset = self.constants
        if isinstance(self.constants, dict):
            energy_offset = sum(self.constant.values())

        # Same as frozen core energy contribution. Note that we only sum over occupied orbitals
        h_pq_energy = 2 * np.sum(h_pq.diagonal() * self.occupations)

        h_pqrs_energy = 0
        for i in range(self.nelec):
            for j in range(self.nelec):
                # Hartree term and exchange term
                h_pqrs_energy += (2 * h_pqrs[i, j, j, i] - h_pqrs[i, j, i, j]) * self.occupations[i] * self.occupations[j]

        return (h_pq_energy + h_pqrs_energy + energy_offset).real

    def hf_energy_qiskit(self, mapper: FermionicMapper | None = JordanWignerMapper()):
        """Calculate the mean-field energy of the single Slater determinant given by self.occupations.
        If this Slater determinant is the Hartree-Fock state, the energy is the Hartree-Fock energy.

        The energy is calculated using Qiskit by using its HartreeFock class and statevector simulation.

        The calculated energy is:
        E = 2 \sum_i h_{ii} + \sum_{ij) (2 h_{ijji} - h_{ijij}) + E_{const.}
        where the sums go over the occupied orbitals

        Args:
            mapper (FermionicMapper | None, optional): _description_. Defaults to JordanWignerMapper().

        Returns:
            float: The calculated mean-field energy
        """
        problem = self.to_qiskit_problem()
        problem.reference_energy = self.fci_energy

        if mapper is None:
            mapper = JordanWignerMapper()
        energy = problem.hamiltonian
        fermionic_op = energy.second_q_op()
        pauli_sum_op = mapper.map(fermionic_op)

        logging.info("pauli_sum_op.num_qubits: %i", pauli_sum_op.num_qubits)

        # Qubit mapping
        logging.info("Using Hartree-Fock state")
        initial_state = HartreeFock(
            num_spatial_orbitals=problem.num_spatial_orbitals,
            num_particles=problem.num_particles,
            qubit_mapper=mapper,
        )
        # estimator = Estimator()
        estimator = StatevectorEstimator()

        hf_energy = estimator.run([(initial_state, pauli_sum_op)]).result()[0].data.evs
        hf_energy += sum((energy.constants.values())).real
        logging.info("HF-result: %f", hf_energy)

        return hf_energy

    def fci_statevector(self) -> Statevector:
        """Return the statevector of the FCI state

        Returns:
            Statevector: Qiskit statevector object of the FCI state
        """
        if len(self.fci_evcs) == 0:
            self.solve_fci()
        # TODO: Not optimal to calculate nelec manually. Add this to Wfc class?
        nelec = (self.nelec, self.nelec)
        norb = self.norb
        fci_state = dopyqo.fci_vector_matrix.statevector_from_fci(fci_vector=self.fci_evcs[0], nelec=nelec, norb=norb)
        return fci_state

    def vqe_statevector(self, vqe_params: np.ndarray | None = None) -> Statevector:
        """Return the statevector of the state prepared by the TenCirChem VQE ansatz

        Args:
            vqe_params (np.ndarray | None, optional): Parameters used for the VQE ansatz. If None, the parameters determined of a prior VQE optimization are used. Defaults to None.

        Raises:
            ValueError: If no TenCirChem VQE optimization has been performed and no parameters are given.

        Returns:
            Statevector: Qiskit statevector object of the state prepared by the TenCirChem VQE ansatz
        """
        if self.tcc_vqe_result is not None or vqe_params is not None:
            logging.info("Calculating VQE statevector from TenCirChem result!")
            ci_vec = self.tcc_ansatz.civector(self.tcc_vqe_params if vqe_params is None else vqe_params)
            nelec = (
                self.nelec,
                self.nelec,
            )
            # state = fci_vector_matrix.statevector_from_civector_restricted(ci_vec, nelec, self.norb)

            dim_spin = np.sqrt(ci_vec.size)
            if not np.isclose(int(dim_spin), dim_spin):
                print(
                    f"{RED}Statevector error: Size of TenCirChem statevector ({ci_vec.size}) is not n^2 (n={dim_spin}) where n should be a squared integer. This should not happen. Please, contact a developer!{RESET_COLOR}"
                )
                sys.exit(1)
            dim_spin = int(dim_spin)
            new_shape = (dim_spin,) * 2
            state = fci_vector_matrix.statevector_from_fci(ci_vec.reshape(new_shape), nelec, self.norb)
        else:
            raise ValueError("No TenCirChem VQE optimization has been performed and no parameters are given. Cannot calculate VQE statevector!")

        return state

    def vqe_statevector_qiskit(self, vqe_params: np.ndarray | None = None) -> Statevector:
        """Return the statevector of the state prepared by the Qiskit VQE ansatz

        Args:
            vqe_params (np.ndarray | None, optional): Parameters used for the VQE ansatz. If None, the parameters determined of a prior VQE optimization are used. Defaults to None.

        Raises:
            ValueError: If no Qiskit VQE optimization has been performed and no parameters are given.

        Returns:
            Statevector: Qiskit statevector object of the state prepared by the Qiskit VQE ansatz
        """
        if self.qiskit_vqe_result is not None or vqe_params is not None:
            logging.info("Calculating VQE statevector from qiskit result!")
            qc = self.qiskit_ansatz
            if vqe_params is None:
                vqe_params = self.qiskit_vqe_result.optimal_parameters
            else:
                # param_vec_elements = ParameterVector("t", length=len(vqe_params)).params
                # vqe_params = {vec_elem: val for vec_elem, val in zip(param_vec_elements, vqe_params)}
                vqe_params = vqe_params

            bc = qc.assign_parameters(vqe_params)

            n_qbits = self.qiskit_pauli_sum_op.num_qubits

            state = Statevector.from_int(i=0, dims=2**n_qbits)
            state = state.evolve(bc)
        else:
            raise ValueError("No Qiskit VQE optimization has been performed and no parameters are given. Cannot calculate VQE statevector!")

        return state

    def run_qiskit_ansatz(
        self,
        parameters: np.ndarray,
        UCCSD_reps: int = 1,
        mapper: FermionicMapper | None = JordanWignerMapper(),
    ) -> float:
        """Run the Qiskit UCCSD ansatz and return the estimated energy

        Args:
            parameters (np.ndarray): Parameters used for the ansatz.
            UCCSD_reps (int, optional): Number of times the UCCSD ansatz is repeated with new parameters. Defaults to 1.
            mapper (FermionicMapper | None, optional): Used Fermion-to-qubit mapping. Defaults to JordanWignerMapper().

        Returns:
            float: The energy of the ansatz
        """
        problem = self.to_qiskit_problem()

        if mapper is None:
            mapper = JordanWignerMapper()
        energy = problem.hamiltonian
        energy_offset = np.sum(list(energy.constants.values()))
        fermionic_op = energy.second_q_op()
        pauli_sum_op = mapper.map(fermionic_op)

        logging.info("pauli_sum_op.num_qubits: %f", pauli_sum_op.num_qubits)

        # Qubit mapping
        logging.info("Using Hartree-Fock state")
        initial_state = HartreeFock(
            num_spatial_orbitals=problem.num_spatial_orbitals,
            num_particles=problem.num_particles,
            qubit_mapper=mapper,
        )

        ansatz = UCCSD(
            problem.num_spatial_orbitals,
            problem.num_particles,
            mapper,
            initial_state=initial_state,
            reps=UCCSD_reps,
        )

        estimator = Estimator()

        job = estimator.run([ansatz], [pauli_sum_op], [parameters])
        estimator_result = job.result().values  #  + energy_offset

        self.qiskit_vqe_initial_state = initial_state
        self.qiskit_problem = problem
        self.qiskit_ansatz = ansatz
        self.qiskit_pauli_sum_op = pauli_sum_op

        return estimator_result

    def run_tcc_ansatz(
        self, parameters: np.ndarray, UCCSD_reps: int = 1, excitations: list[tuple[int, ...]] | None = None, qiskit_equivalent: bool = False
    ) -> float:
        """
        Run the TenCirChem UCCSD ansatz and return the estimated energy

        Args:
            parameters (np.ndarray): Parameters used for the VQE ansatz.
            UCCSD_reps (int, optional): Number of times the UCCSD ansatz is repeated with new parameters. Defaults to 1.
            excitations (list[tuple[int, ...]] | None, optional): Excitations used in the ansatz. If None all double and single excitations are used. Defaults to None.
            qiskit_equivalent (bool, optional): If set to True, the same excitations and ordering as in Qiskit is used.
                                                With this, the provided parameters give the same result as the run_qiskit_ansatz function. Defaults to False.

        Returns:
            float: The energy of the ansatz
        """
        # NOTE: TenCirChem can only simulate spin-restricted systems!
        nelec = (
            self.nelec,
            self.nelec,
        )

        energy_offset = self.constants
        if isinstance(self.constants, dict):
            energy_offset = sum(self.constant.values())

        one_mo = self.h_pq
        two_mo = self.h_pqrs.swapaxes(1, 2).swapaxes(1, 3)

        assert np.allclose(one_mo.real, one_mo), "Single electron matrix elements are complex. TenCirChem only supports real matrix elements!"
        assert np.allclose(two_mo.real, two_mo), "Two electron matrix elements are complex. TenCirChem only supports real matrix elements!"
        assert np.allclose(energy_offset.real, energy_offset), "Constant energy offset is complex. TenCirChem only supports real offsets!"
        one_mo = one_mo.real
        two_mo = two_mo.real
        constant = energy_offset.real

        kwargs = {"mo_coeff": np.eye(self.norb), "init_method": "zeros", "run_hf": False, "run_mp2": False, "run_ccsd": False, "run_fci": False}
        tcc_uccsd = tencirchem.UCCSD.from_integral(one_mo, two_mo, nelec, constant, **kwargs)
        # See https://tensorcircuit.github.io/TenCirChem-NG/faq.html#why-are-the-number-of-excitation-operators-and-the-number-of-parameters-different-what-is-param-ids
        tcc_uccsd.param_ids = None
        if excitations is not None:
            tcc_uccsd.ex_ops = excitations
        else:
            singles_tcc = tcc_helpers.get_ex1_ops(self.norb, nelec)
            doubles_tcc = tcc_helpers.get_ex2_ops(self.norb, nelec)
            # NOTE: Sort indices of excitaions like qiskit
            singles_tcc_sorted = sorted(singles_tcc)
            doubles_tcc_sorted = sorted([(*sorted(x[:2]), *sorted(x[2:])) for x in doubles_tcc])
            tcc_uccsd.ex_ops = doubles_tcc_sorted + singles_tcc_sorted

            logging.info("TenCirChem UCCSD has %s single and %s doubles.", len(singles_tcc_sorted), len(doubles_tcc_sorted))

            if qiskit_equivalent:
                #####################################################################################
                ##                                  Qiskit UCCSD                                   ##
                #####################################################################################
                integrals = ElectronicIntegrals.from_raw_integrals(one_mo, two_mo, auto_index_order=True)
                qiskit_energy = ElectronicEnergy(integrals)
                qiskit_problem = ElectronicStructureProblem(qiskit_energy)
                qiskit_problem.num_particles = nelec
                qiskit_problem.num_spatial_orbitals = self.norb
                mapper = JordanWignerMapper()
                initial_state = HartreeFock(
                    num_spatial_orbitals=qiskit_problem.num_spatial_orbitals,
                    num_particles=qiskit_problem.num_particles,
                    qubit_mapper=mapper,
                )
                qiskit_uccsd = UCCSD(
                    qiskit_problem.num_spatial_orbitals,
                    qiskit_problem.num_particles,
                    mapper,
                    initial_state=initial_state,
                    reps=1,
                )

                #####################################################################################
                ##               Making TenCirChem UCCSD equivalent to Qiskit UCCSD                ##
                #####################################################################################
                # The indices of the creation/annihilation operators in the double excitations are sorted in qiskit.
                # We do the same for the TenCirChem excitations
                doubles_tcc_sorted = [(*sorted(x[:2]), *(sorted(x[2:]))) for x in doubles_tcc]
                tcc_uccsd.ex_ops = singles_tcc + doubles_tcc_sorted

                # Bringing the qiskit excitations into the same notation as now used in tcc_uccsd.ex_ops to make them comparable
                qiskit_ex_list = [
                    (x[1][0], x[0][0]) if len(x[0]) == 1 else (x[1][0], x[1][1], x[0][0], x[0][1]) for x in qiskit_uccsd.excitation_list
                ]

                # We map the indices of qiskit excitations to indices of TenCirChem excitations
                tcc_idc = [tcc_uccsd.ex_ops.index(ex_qiskit) for ex_qiskit in qiskit_ex_list]

                # With this mapping we now reorder the excitations in the TenCirChem UCCSD to have the same ordering as in the qiskit UCCSD
                tcc_uccsd.ex_ops = [tcc_uccsd.ex_ops[idx] for idx in tcc_idc]

            tcc_uccsd.ex_ops *= UCCSD_reps

        #####################################################################################
        ##                                 Running Ansatz                                  ##
        #####################################################################################
        assert len(parameters) == tcc_uccsd.n_params, f"TenCirChem ansatz expects {tcc_uccsd.n_params} parameters but {len(parameters)} were given!"
        tcc_energy = tcc_uccsd.energy(parameters)

        self.tcc_ansatz = tcc_uccsd

        return tcc_energy

    def solve_vqe(
        self,
        optimizer: dopyqo.VQEOptimizers = dopyqo.VQEOptimizers.L_BFGS_B,
        UCCSD_reps: int = 1,
        qiskit_equivalent: bool = False,
        initial_params: np.ndarray | None = None,
        maxiter: int | None = None,
        excitations: list[tuple[int, ...]] | dopyqo.ExcitationPools = dopyqo.ExcitationPools.SINGLES_DOUBLES,
    ) -> scipy.optimize.OptimizeResult:
        """Run a VQE optimization with TenCirChem and a UCCSD ansatz. The initial state, that is prepared before the parametrized ansatz is applied,
        is a Slater determinant matching self.occupations.

        Args:
            optimizer (dopyqo.VQEOptimizers, optional): Classical optimizer used in the VQE. Defaults to dopyqo.VQEOptimizers.L_BFGS_B.
            UCCSD_reps (int, optional): Number of times the UCCSD ansatz is repeated with new parameters. Defaults to 1.
            qiskit_equivalent (bool, optional): If set to True, the same excitations and ordering as in Qiskit is used.
                                                With this, the provided parameters give the same result as the solve_vqe_qiskit function. Defaults to False.
            initial_params (np.ndarray | None, optional): Parameters to start the VQE optimization with. Defaults to None.
            maxiter (int | None, optional): Maximum number of optimizer iterations before stopping the optimization. If None, large defaults are used. Defaults to None.
            excitations (list[tuple[int, ...]] | dopyqo.ExcitationPools, optional): Excitations used in the ansatz. Defaults to dopyqo.ExcitationPools.SINGLES_DOUBLES where all double and single excitations are used.

        Returns:
            scipy.optimize.OptimizeResult: Scipy optimization result
        """
        # tencirchem.set_backend("cupy")
        # NOTE: TenCirChem can only simulate spin-restricted systems!

        nelec = (
            self.nelec,
            self.nelec,
        )

        energy_offset = self.constants
        if isinstance(self.constants, dict):
            energy_offset = sum(self.constant.values())

        one_mo = self.h_pq
        two_mo = self.h_pqrs.swapaxes(1, 2).swapaxes(1, 3)

        if not np.allclose(one_mo.real, one_mo):
            print(
                f"{RED}Hamiltonian error: Single electron matrix elements are complex. TenCirChem only supports real matrix elements!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        if not np.allclose(two_mo.real, two_mo):
            print(
                f"{RED}Hamiltonian error: Two electron matrix elements are complex. TenCirChem only supports real matrix elements!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        if not np.allclose(energy_offset.real, energy_offset):
            print(
                f"{RED}Hamiltonian error: Constant energy offset is complex. TenCirChem only supports real offsets!!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        one_mo = one_mo.real
        two_mo = two_mo.real
        energy_offset = energy_offset.real

        # mo_coeff = np.eye(self.norb) if self.mo_coeff is None else self.mo_coeff
        # print(f"{mo_coeff=}")
        mo_coeff = np.eye(self.norb)
        kwargs = {
            "mo_coeff": mo_coeff,
            "init_method": "zeros",
            "run_hf": False,
            "run_mp2": False,
            "run_ccsd": False,
            "run_fci": False,
            # "engine": "civector",
        }
        tcc_uccsd = tencirchem.UCCSD.from_integral(one_mo, two_mo, nelec, energy_offset, **kwargs)
        # print(f"\t Using TenCirChem engine {tcc_uccsd.engine}")
        # See https://tensorcircuit.github.io/TenCirChem-NG/faq.html#why-are-the-number-of-excitation-operators-and-the-number-of-parameters-different-what-is-param-ids
        tcc_uccsd.param_ids = None
        if isinstance(excitations, list):
            tcc_uccsd.ex_ops = excitations
        else:
            singles_tcc = tcc_helpers.get_ex1_ops(self.norb, nelec)
            doubles_tcc = tcc_helpers.get_ex2_ops(self.norb, nelec)
            # Sort indices of excitations like qiskit
            singles_tcc_sorted = sorted(singles_tcc)
            doubles_tcc_sorted = sorted([(*sorted(x[:2]), *sorted(x[2:])) for x in doubles_tcc])

            match excitations:
                case dopyqo.ExcitationPools.SINGLES:
                    excitation_list = singles_tcc_sorted
                case dopyqo.ExcitationPools.DOUBLES:
                    excitation_list = doubles_tcc_sorted
                case dopyqo.ExcitationPools.SINGLES_DOUBLES:
                    excitation_list = doubles_tcc_sorted + singles_tcc_sorted
                case _:
                    print(
                        f"{RED}VQE error: Parameter {excitations=} is not supported. Use a list of excitation tuples or dopyqo.ExcitationPools types!{RESET_COLOR}"
                    )
                    sys.exit(1)

            tcc_uccsd.ex_ops = excitation_list
            logging.info("TenCirChem UCCSD has %s excitations.", len(excitation_list))

            if qiskit_equivalent:
                #####################################################################################
                ##                                  Qiskit UCCSD                                   ##
                #####################################################################################
                integrals = ElectronicIntegrals.from_raw_integrals(one_mo, two_mo, auto_index_order=True)
                qiskit_energy = ElectronicEnergy(integrals)
                qiskit_problem = ElectronicStructureProblem(qiskit_energy)
                qiskit_problem.num_particles = nelec
                qiskit_problem.num_spatial_orbitals = self.norb
                mapper = JordanWignerMapper()
                initial_state = HartreeFock(
                    num_spatial_orbitals=qiskit_problem.num_spatial_orbitals,
                    num_particles=qiskit_problem.num_particles,
                    qubit_mapper=mapper,
                )
                qiskit_uccsd = UCCSD(
                    qiskit_problem.num_spatial_orbitals,
                    qiskit_problem.num_particles,
                    mapper,
                    initial_state=initial_state,
                    reps=1,
                )
                # Fix UCCSD operator order: first doubles, then singles
                doubles = [op for (op, excitation) in zip(qiskit_uccsd.operators, qiskit_uccsd.excitation_list) if len(excitation[0]) == 2]
                singles = [op for (op, excitation) in zip(qiskit_uccsd.operators, qiskit_uccsd.excitation_list) if len(excitation[0]) == 1]
                qiskit_uccsd.operators = doubles + singles
                qiskit_uccsd._invalidate()

                #####################################################################################
                ##               Making TenCirChem UCCSD equivalent to Qiskit UCCSD                ##
                #####################################################################################
                # Bringing the qiskit excitations into the same notation as now used in tcc_uccsd.ex_ops to make them comparable
                qiskit_ex_list = [
                    (x[1][0], x[0][0]) if len(x[0]) == 1 else (x[1][0], x[1][1], x[0][0], x[0][1]) for x in qiskit_uccsd.excitation_list
                ]

                # We map the indices of qiskit excitations to indices of TenCirChem excitations
                tcc_idc = [tcc_uccsd.ex_ops.index(ex_qiskit) for ex_qiskit in qiskit_ex_list if ex_qiskit in tcc_uccsd.ex_ops]

                # With this mapping we now reorder the excitations in the TenCirChem UCCSD to have the same ordering as in the qiskit UCCSD
                tcc_uccsd.ex_ops = [tcc_uccsd.ex_ops[idx] for idx in tcc_idc]

            tcc_uccsd.ex_ops *= UCCSD_reps

        # Setting tcc_uccsd.init_state to match self.occupations
        binary_string_tmp = "0b" + "".join(map(str, map(int, reversed(self.occupations)))) * 2
        binary_string_tmp = int(binary_string_tmp, base=2)
        ci_strings_tmp = tcc_uccsd.get_ci_strings()  # .get()
        ci_string_idx_tmp = np.where(ci_strings_tmp == binary_string_tmp)[0][0]
        #
        init_statevector_tmp = np.zeros(len(ci_strings_tmp))
        init_statevector_tmp[ci_string_idx_tmp] = 1.0
        tcc_uccsd.init_state = list(init_statevector_tmp)

        #####################################################################################
        ##                                  Running VQE                                    ##
        #####################################################################################
        # Manual optimization
        times_tcc = []
        eval_count_tcc = 0
        counts_tcc = []
        values_tcc = []
        tcc_vqe_params_lst = []

        def cost(x):
            nonlocal times_tcc
            nonlocal eval_count_tcc
            nonlocal counts_tcc
            nonlocal values_tcc

            tcc_energy = tcc_uccsd.energy(x)
            times_tcc.append(time.perf_counter())

            eval_count_tcc += 1
            print(
                f"Optimizer evaluation #{eval_count_tcc}, Diff. to ref.: {np.abs(tcc_energy - self.reference_energy)}",
                end="\r",
                flush=True,
            )
            counts_tcc.append(eval_count_tcc)
            values_tcc.append(tcc_energy)
            tcc_vqe_params_lst.append(x)

            return tcc_energy

        n_params_tmp = tcc_uccsd.n_params
        if initial_params is None:
            initial_params = np.zeros(n_params_tmp)
        else:
            if len(initial_params) != n_params_tmp:
                print(
                    f"{RED}VQE error: initial_params has length {len(initial_params)} but VQE ansatz has {n_params_tmp} parameters!{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
        logging.info("Using optimizer %s", optimizer)

        match optimizer:
            case dopyqo.VQEOptimizers.COBYLA:
                maxiter = 1e6 if maxiter is None else maxiter
                options = dict(maxiter=maxiter, tol=1e-10)
                optimizer_func = "COBYLA"
            case dopyqo.VQEOptimizers.L_BFGS_B:
                maxiter = 1e6 if maxiter is None else maxiter
                options = dict(maxfun=maxiter * 4 * tcc_uccsd.n_params, maxiter=maxiter, ftol=1e-13)
                optimizer_func = "L-BFGS-B"
            case dopyqo.VQEOptimizers.ExcitationSolve:
                from excitationsolve import ExcitationSolveScipy

                maxiter = 100 if maxiter is None else maxiter
                excsolve_obj = ExcitationSolveScipy(maxiter=maxiter, tol=1e-10, save_parameters=True)
                optimizer_func = excsolve_obj.minimize
                options = dict(print_instead_logging=False, reference_energy=self.reference_energy)
            case _:
                print(f"{RED}VQE error: Parameter {optimizer=} is not supported. Use dopyqo.VQEOptimizers types!{RESET_COLOR}")
                sys.exit(1)
        res_tcc = scipy.optimize.minimize(cost, initial_params, method=optimizer_func, options=options)
        params_tcc = res_tcc.x

        match optimizer:
            case dopyqo.VQEOptimizers.ExcitationSolve:
                counts_tcc = excsolve_obj.nfevs
                values_tcc = excsolve_obj.energies

        self.tcc_vqe_params = params_tcc
        self.tcc_vqe_params_lst = tcc_vqe_params_lst
        # To have the same parameter values in the qiskit UCCSD circuit we have to flip the sign
        if qiskit_equivalent:
            self.qiskit_vqe_params = -params_tcc

        self.tcc_vqe_counts = np.array(counts_tcc)
        self.tcc_vqe_values = np.array(values_tcc)

        self.tcc_vqe_result = res_tcc
        self.tcc_ansatz = tcc_uccsd

        self.vqe_counts = self.tcc_vqe_counts
        self.vqe_values = self.tcc_vqe_values
        self.vqe_result = self.tcc_vqe_result

        return self.tcc_vqe_result

    def solve_vqe_qiskit(
        self,
        optimizer: dopyqo.VQEOptimizers = dopyqo.VQEOptimizers.L_BFGS_B,
        UCCSD_reps: int = 1,
        mapper: FermionicMapper | None = JordanWignerMapper(),
        maxiter: int | None = None,
        excitations: list[tuple[int, ...]] | dopyqo.ExcitationPools = dopyqo.ExcitationPools.SINGLES_DOUBLES,
    ) -> VQEResult:
        """Run a VQE optimization with Qiskit and a UCCSD ansatz. The initial state, that is prepared before the parametrized ansatz is applied,
        is a Slater determinant matching self.occupations.

        Args:
            mapper (FermionicMapper | None, optional): Used Fermion-to-qubit mapping. Defaults to JordanWignerMapper().
            optimizer (dopyqo.VQEOptimizers, optional): Classical optimizer used in the VQE. Defaults to dopyqo.VQEOptimizers.L_BFGS_B.
            maxiter (int | None, optional): Maximum number of optimizer iterations before stopping the optimization. If None, large defaults are used. Defaults to None.
            UCCSD_reps (int, optional): Number of times the UCCSD ansatz is repeated with new parameters. Defaults to 1.

        Returns:
            VQEResult: Qiskit VQEResult object
        """
        energy_offset = self.constants
        if isinstance(self.constants, dict):
            energy_offset = sum(self.constant.values())

        one_mo = self.h_pq
        two_mo = self.h_pqrs.swapaxes(1, 2).swapaxes(1, 3)
        # if not np.allclose(one_mo.real, one_mo):
        #     print(
        #         f"{RED}Hamiltonian error: Single electron matrix elements are complex. Qiskit only supports real matrix elements!{RESET_COLOR}",
        #         file=sys.stderr,
        #     )
        #     sys.exit(1)
        # if not np.allclose(two_mo.real, two_mo):
        #     print(
        #         f"{RED}Hamiltonian error: Two electron matrix elements are complex. Qiskit only supports real matrix elements!{RESET_COLOR}",
        #         file=sys.stderr,
        #     )
        #     sys.exit(1)
        auto_index_order = True
        if not np.allclose(one_mo.real, one_mo):
            # print(
            #     f"{ORANGE}Hamiltonian error: Single electron matrix elements are complex. Qiskit only supports real matrix elements!{RESET_COLOR}",
            # )
            auto_index_order = False
        if not np.allclose(two_mo.real, two_mo):
            # print(
            #     f"{ORANGE}Hamiltonian error: Two electron matrix elements are complex. Qiskit only supports real matrix elements!{RESET_COLOR}",
            # )
            auto_index_order = False
        if not np.allclose(energy_offset.real, energy_offset):
            print(
                f"{RED}Hamiltonian error: Constant energy offset is complex. Qiskit only supports real offsets!!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)

        problem = self.to_qiskit_problem(auto_index_order=auto_index_order)
        problem.reference_energy = self.fci_energy

        if mapper is None:
            mapper = JordanWignerMapper()
        energy = problem.hamiltonian
        fermionic_op = energy.second_q_op()
        pauli_sum_op = mapper.map(fermionic_op)

        logging.info("pauli_sum_op.num_qubits: %f", pauli_sum_op.num_qubits)

        # Qubit mapping
        logging.info("Using Hartree-Fock state")
        initial_state = HartreeFock(
            num_spatial_orbitals=problem.num_spatial_orbitals,
            num_particles=problem.num_particles,
            qubit_mapper=mapper,
        )
        # print("Initial state:")
        # print(initial_state.draw())
        # from qiskit.quantum_info import Statevector

        # print(Statevector(initial_state).draw("latex_source"))
        ansatz = UCCSD(
            problem.num_spatial_orbitals,
            problem.num_particles,
            mapper,
            initial_state=initial_state,
            reps=UCCSD_reps,
        )

        doubles = [op for (op, excitation) in zip(ansatz.operators, ansatz.excitation_list) if len(excitation[0]) == 2]
        singles = [op for (op, excitation) in zip(ansatz.operators, ansatz.excitation_list) if len(excitation[0]) == 1]
        match excitations:
            case dopyqo.ExcitationPools.SINGLES:
                ansatz.operators = singles
            case dopyqo.ExcitationPools.DOUBLES:
                ansatz.operators = doubles
            case dopyqo.ExcitationPools.SINGLES_DOUBLES:
                ansatz.operators = doubles + singles
            case _:
                print(f"{RED}VQE error: Parameter {excitations=} is not supported. Use dopyqo.ExcitationPools types!{RESET_COLOR}")
                sys.exit(1)
        ansatz._invalidate()

        logging.info("ansatz.num_qubits: %i", ansatz.num_qubits)

        # initial_state.draw(
        #     "mpl", filename=os.path.join("results", "UCCSD_initial_state")
        # )

        # ansatz.draw("mpl", filename=os.path.join("results", "vqe_ansatz"))

        match optimizer:
            case dopyqo.VQEOptimizers.COBYLA:
                maxiter = 1e6 if maxiter is None else maxiter
                optimizer_obj = optim.COBYLA(maxiter=maxiter, tol=1e-13)
            case dopyqo.VQEOptimizers.L_BFGS_B:
                maxiter = 1e6 if maxiter is None else maxiter
                optimizer_obj = optim.L_BFGS_B(maxfun=maxiter * 4 * len(ansatz.operators), maxiter=maxiter, ftol=1e-13)
            case dopyqo.VQEOptimizers.ExcitationSolve:
                from excitationsolve.excitation_solve_qiskit import ExcitationSolveQiskit

                maxiter = 100 if maxiter is None else maxiter
                optimizer_obj = ExcitationSolveQiskit(maxiter=maxiter, tol=1e-12)
            case _:
                print(f"{RED}VQE error: Parameter {optimizer=} is not supported. Use dopyqo.VQEOptimizers types!{RESET_COLOR}")
                sys.exit(1)
        estimator = Estimator()
        # estimator = StatevectorEstimator()

        # logging.info("HF-result: %s", estimator.run(initial_state, pauli_sum_op).result().values)

        offset = np.sum(list(energy.constants.values()))
        counts = []
        values = []

        def store_intermediate_result(eval_count, parameters, mean, std):
            print(
                f"Optimizer evaluation #{eval_count}, Diff. to ref.: {np.abs(mean + offset - self.reference_energy)}",
                end="\r",
            )
            counts.append(eval_count)
            values.append(mean)

        solver = VQE(
            estimator,
            ansatz,
            optimizer_obj,
            callback=store_intermediate_result,
            initial_point=np.zeros((ansatz.num_parameters,)),
        )
        logging.info("VQE defined")
        logging.info("Solving VQE...")
        vqe_result = solver.compute_minimum_eigenvalue(pauli_sum_op)
        logging.info("Solved")
        elec_struc_result = problem.interpret(vqe_result)

        match optimizer:
            case dopyqo.VQEOptimizers.ExcitationSolve:
                counts = optimizer_obj.nfevs
                values = optimizer_obj.energies

        self.qiskit_vqe_solver = solver
        self.qiskit_vqe_result = vqe_result

        self.qiskit_vqe_initial_state = initial_state
        self.qiskit_vqe_ansatz = solver.ansatz

        self.qiskit_vqe_counts = np.array(counts)
        self.qiskit_vqe_values = np.array(values) + offset

        self.qiskit_elec_struc_result = elec_struc_result
        self.qiskit_vqe_result = vqe_result
        self.qiskit_problem = problem
        self.qiskit_ansatz = ansatz
        self.qiskit_pauli_sum_op = pauli_sum_op

        self.vqe_counts = self.qiskit_vqe_counts
        self.vqe_values = self.qiskit_vqe_values
        self.vqe_result = self.qiskit_elec_struc_result

        return self.qiskit_vqe_result

    def solve_vqe_adapt(
        self,
        optimizer: dopyqo.VQEOptimizers = dopyqo.VQEOptimizers.L_BFGS_B,
        maxiter: int | None = None,
        excitation_pool: list[tuple[int, ...]] | dopyqo.ExcitationPools = dopyqo.ExcitationPools.SINGLES_DOUBLES,
        drain_pool: bool = True,
        conv_threshold: float = 1e-13,
        selection_criterion: dopyqo.AdaptSelectionCriterion = dopyqo.AdaptSelectionCriterion.ENERGY,
    ) -> scipy.optimize.OptimizeResult:
        """Run a ADAPT-VQE optimization with TenCirChem given a pool of excitations. The initial state,
        that is prepared before the parametrized ansatz is applied, is a Slater determinant matching self.occupations.
        All appended excitations are optimized after every appended excitation.
        The criterion with which a excitation is selected from the pool can be set with the selection_criterion parameter.

        Args:
            optimizer (dopyqo.VQEOptimizers, optional): Classical optimizer used in the VQE. Defaults to dopyqo.VQEOptimizers.L_BFGS_B.
            maxiter (int | None, optional): Maximum number of optimizer iterations used for optimizing all appended excitations before
                                            stopping the optimization. If None, large defaults are used. Defaults to None.
            excitation_pool (list[tuple[int, ...]] | dopyqo.ExcitationPools, optional): Excitations in the ADAPT pool. Defaults to
                                                                                        dopyqo.ExcitationPools.SINGLES_DOUBLES where all double and
                                                                                        single excitations are in the pool.
            drain_pool (bool, optional): If set to True an excitation is removed from the pool if it was appended to the ansatz. This means each
                                         operator can only appear once in the ansatz. Defaults to True.
            conv_threshold (float, optional): Excitations are appended to the ansatz until their initial impact drops below this threshold value.
                                              Defaults to 1e-13.
            selection_criterion (dopyqo.AdaptSelectionCriterion, optional): Used operator selection criterion.
                                                                            When dopyqo.AdaptSelectionCriterion.ENERGY is used, every appended
                                                                            excitation is initialized with its optimal parameter.Defaults to
                                                                            dopyqo.AdaptSelectionCriterion.ENERGY.

        Returns:
            scipy.optimize.OptimizeResult: Scipy optimization result
        """
        # NOTE: TenCirChem can only simulate spin-restricted systems!

        nelec = (
            self.nelec,
            self.nelec,
        )

        energy_offset = self.constants
        if isinstance(self.constants, dict):
            energy_offset = sum(self.constant.values())

        one_mo = self.h_pq
        two_mo = self.h_pqrs.swapaxes(1, 2).swapaxes(1, 3)  # To TenCirChem/PySCF notation

        if not np.allclose(one_mo.real, one_mo):
            print(
                f"{RED}Hamiltonian error: Single electron matrix elements are complex. TenCirChem only supports real matrix elements!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        if not np.allclose(two_mo.real, two_mo):
            print(
                f"{RED}Hamiltonian error: Two electron matrix elements are complex. TenCirChem only supports real matrix elements!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        if not np.allclose(energy_offset.real, energy_offset):
            print(
                f"{RED}Hamiltonian error: Constant energy offset is complex. TenCirChem only supports real offsets!!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        one_mo = one_mo.real
        two_mo = two_mo.real
        energy_offset = energy_offset.real

        # mo_coeff = np.eye(self.norb) if self.mo_coeff is None else self.mo_coeff
        # print(f"{mo_coeff=}")
        mo_coeff = np.eye(self.norb)
        kwargs = {
            "mo_coeff": mo_coeff,
            "init_method": "zeros",
            "run_hf": False,
            "run_mp2": False,
            "run_ccsd": False,
            "run_fci": False,
            # "engine": "civector",
        }
        ansatz = tencirchem.UCCSD.from_integral(one_mo, two_mo, nelec, energy_offset, **kwargs)
        print(f"{ansatz.engine=}")
        # See https://tensorcircuit.github.io/TenCirChem-NG/faq.html#why-are-the-number-of-excitation-operators-and-the-number-of-parameters-different-what-is-param-ids
        ansatz.param_ids = None
        ansatz.ex_ops = []

        if not isinstance(excitation_pool, list):
            singles_tcc = tcc_helpers.get_ex1_ops(self.norb, nelec)
            doubles_tcc = tcc_helpers.get_ex2_ops(self.norb, nelec)
            # NOTE: Sort indices of excitaions like qiskit
            singles_tcc_sorted = sorted(singles_tcc)
            doubles_tcc_sorted = sorted([(*sorted(x[:2]), *sorted(x[2:])) for x in doubles_tcc])

            match excitation_pool:
                case dopyqo.ExcitationPools.SINGLES:
                    excitation_pool = singles_tcc_sorted
                case dopyqo.ExcitationPools.DOUBLES:
                    excitation_pool = doubles_tcc_sorted
                case dopyqo.ExcitationPools.SINGLES_DOUBLES:
                    excitation_pool = singles_tcc_sorted + doubles_tcc_sorted
                case _:
                    print(
                        f"{RED}ADAPT-VQE error: Parameter {excitation_pool=} is not supported. Use a list of excitation tuples or dopyqo.ExcitationPools types!{RESET_COLOR}"
                    )
                    sys.exit(1)

        # Setting tcc_uccsd.init_state to match self.occupations
        binary_string_tmp = "0b" + "".join(map(str, map(int, reversed(self.occupations)))) * 2
        binary_string_tmp = int(binary_string_tmp, base=2)
        ci_strings_tmp = ansatz.get_ci_strings()
        ci_string_idx_tmp = np.where(ci_strings_tmp == binary_string_tmp)[0][0]
        #
        init_statevector_tmp = np.zeros(len(ci_strings_tmp))
        init_statevector_tmp[ci_string_idx_tmp] = 1.0
        ansatz.init_state = list(init_statevector_tmp)

        #####################################################################################
        ##                           Cost function on parameter                            ##
        #####################################################################################
        times_tcc = []
        eval_count_tcc = 0
        counts_tcc = []
        values_tcc = []
        tcc_vqe_params_lst = []

        # Tracking values and counts has to be handled differently for ExcSolve,
        # since we want to manually set the values and counts from the ExcSolve object
        eval_count_tcc_excsolve = 0
        counts_tcc_excsolve = []
        values_tcc_excsolve = []
        tcc_vqe_params_lst_excsolve = []

        def cost_one_exc(x):
            nonlocal times_tcc
            nonlocal eval_count_tcc
            nonlocal counts_tcc
            nonlocal values_tcc
            nonlocal current_params

            tcc_energy = ansatz.energy(np.append(current_params, x))
            times_tcc.append(time.perf_counter())

            eval_count_tcc += 1
            print(
                f"Optimizer evaluation #{eval_count_tcc}, Diff. to ref.: {np.abs(tcc_energy - self.reference_energy)}",
                end="\r",
                flush=True,
            )
            counts_tcc.append(eval_count_tcc)
            values_tcc.append(tcc_energy)
            tcc_vqe_params_lst.append(x)

            return tcc_energy

        #####################################################################################
        ##                        Cost function for all parameters                         ##
        #####################################################################################
        def cost(x):
            nonlocal times_tcc
            nonlocal eval_count_tcc
            nonlocal counts_tcc
            nonlocal values_tcc

            tcc_energy = ansatz.energy(x)
            times_tcc.append(time.perf_counter())

            eval_count_tcc += 1
            print(
                f"Optimizer evaluation #{eval_count_tcc}, Diff. to ref.: {np.abs(tcc_energy - self.reference_energy)}",
                end="\r",
                flush=True,
            )
            counts_tcc.append(eval_count_tcc)
            values_tcc.append(tcc_energy)
            tcc_vqe_params_lst.append(x)

            return tcc_energy

        #####################################################################################
        ##                               Running ADAPT-VQE                                 ##
        #####################################################################################
        converged = False
        pool_drained = False
        current_params = np.zeros((0,))
        initial_param = 0.0
        current_energy = self.hf_energy()
        match selection_criterion:
            case dopyqo.AdaptSelectionCriterion.ENERGY:
                try:
                    from excitationsolve import ExcitationSolveScipy
                except ImportError:
                    print(
                        f"{ORANGE}ADAPT error: Could not import excitationsolve package which is needed for the energy-based operator selection. ",
                        f"Please install the excitationsolve package.{RESET_COLOR}",
                    )
                    sys.exit(1)

        while len(excitation_pool) > 0:
            # Check every excitation in the pool and compute their impact on the energy
            impact_and_param_per_excitation = {}
            current_ansatz_excs = ansatz.ex_ops.copy()
            for exc in excitation_pool:
                ansatz.ex_ops = current_ansatz_excs + [exc]

                match selection_criterion:
                    case dopyqo.AdaptSelectionCriterion.ENERGY:
                        # Using ExcitationSolve for operator selection
                        excsolve_obj = ExcitationSolveScipy(maxiter=1, tol=1e-10, save_parameters=True)
                        optimizer_func = excsolve_obj.minimize
                        options = dict(print_instead_logging=False, reference_energy=self.reference_energy)
                        res_tcc = scipy.optimize.minimize(cost_one_exc, initial_param, method=optimizer_func, options=options)
                        impact_and_param_per_excitation[exc] = (np.abs(current_energy - res_tcc.fun), res_tcc.x)
                        #
                        eval_count_tcc_excsolve += excsolve_obj.nfevs[-1]
                        counts_tcc_excsolve.append(eval_count_tcc_excsolve)
                        values_tcc_excsolve.append(res_tcc.fun)
                        tcc_vqe_params_lst_excsolve.append(res_tcc.x)
                        counts_tcc = counts_tcc[: -excsolve_obj.nfevs[-1]]
                        counts_tcc.append(eval_count_tcc)
                        values_tcc = values_tcc[: -excsolve_obj.nfevs[-1]]
                        values_tcc.append(res_tcc.fun)
                        tcc_vqe_params_lst = tcc_vqe_params_lst[: -excsolve_obj.nfevs[-1]]
                        tcc_vqe_params_lst.append(res_tcc.x)
                    case dopyqo.AdaptSelectionCriterion.GRADIENT:
                        # Compute d/dθ <ψ(θ)|H|ψ(θ)> at θ=0: Equivalent to <ψ(θ)|[H, A]|ψ(θ)> where A is the excitation exc
                        _nrg, grad = ansatz.energy_and_grad(np.append(current_params, 0.0))
                        grad = grad[-1]
                        impact_and_param_per_excitation[exc] = (np.abs(grad), 0.0)

                        for _ in range(4):  # 4 energy evaluations per gradient/s
                            eval_count_tcc += 1
                            counts_tcc.append(eval_count_tcc)
                            values_tcc.append(current_energy)
                    case _:
                        print(
                            f"{RED}ADAPT-VQE error: Parameter {selection_criterion=} is not supported. Use dopyqo.AdaptSelectionCriterion types!{RESET_COLOR}"
                        )
                        sys.exit(1)

            # Append operator with largest impact
            sorted_exc = sorted(impact_and_param_per_excitation.items(), key=lambda x: x[1][0], reverse=True)
            exc_to_append, imp_param_tmp = sorted_exc[0]
            # Stop if impact is less than threshold
            if imp_param_tmp[0] < conv_threshold:
                converged = True
                ansatz.ex_ops = current_ansatz_excs
                print(f"{GREEN}ADAPT-VQE converged!{RESET_COLOR}")
                break
            ansatz.ex_ops = current_ansatz_excs + [exc_to_append]
            if drain_pool:
                excitation_pool = [exc for exc in excitation_pool if exc != exc_to_append]
            current_params = np.append(current_params, imp_param_tmp[1])

            # Optimize all parameters
            logging.info("Using optimizer %s", optimizer)
            match optimizer:
                case dopyqo.VQEOptimizers.COBYLA:
                    maxiter = 1e6 if maxiter is None else maxiter
                    options = dict(maxiter=maxiter, tol=1e-10)
                    optimizer_func = "COBYLA"
                case dopyqo.VQEOptimizers.L_BFGS_B:
                    maxiter = 1e6 if maxiter is None else maxiter
                    options = dict(maxfun=maxiter * 4 * ansatz.n_params, maxiter=maxiter, ftol=1e-13)
                    optimizer_func = "L-BFGS-B"
                case dopyqo.VQEOptimizers.ExcitationSolve:
                    # TODO: If ExcSolve is used and there is only one operator in the ansatz,
                    #       we do not need to optimize here since ExcSolve already initialized the parameter in its optimal value
                    maxiter = 100 if maxiter is None else maxiter
                    excsolve_obj = ExcitationSolveScipy(maxiter=maxiter, tol=1e-10, save_parameters=True)
                    optimizer_func = excsolve_obj.minimize
                    options = dict()
                case _:
                    print(f"{RED}ADAPT-VQE error: Parameter {optimizer=} is not supported. Use dopyqo.VQEOptimizers types!{RESET_COLOR}")
                    sys.exit(1)
            res_tcc = scipy.optimize.minimize(cost, current_params, method=optimizer_func, options=options)
            match optimizer:
                case dopyqo.VQEOptimizers.ExcitationSolve:
                    counts_tcc_excsolve.extend(np.array(excsolve_obj.nfevs) + eval_count_tcc_excsolve)
                    eval_count_tcc_excsolve += excsolve_obj.nfevs[-1]
                    values_tcc_excsolve.extend(excsolve_obj.energies)
                    tcc_vqe_params_lst_excsolve.extend(excsolve_obj.params)

            current_params = res_tcc.x.copy()
            current_energy = res_tcc.fun

            ops_left_str = f" ({len(excitation_pool)} operators left in pool)" if drain_pool else ""
            # print(
            #     f"Done optimizing all {len(current_params)} parameters in the ansatz{ops_left_str}. Initial energy impact of last appended operator ({exc_to_append}): {imp_param_tmp[0]:.2e} (threshold: {conv_threshold}). Current diff. to ref.: {(np.abs(self.reference_energy - res_tcc.fun)):.2e}\n"
            # )
        else:  # Pool is drained. Not executed if break was used to exit while loop
            pool_drained = True
            print(f"{ORANGE}ADAPT-VQE pool is drained!{RESET_COLOR}")

        print(f"{GREEN}ADAPT ansatz consits of {ansatz.n_params} operators!{RESET_COLOR}")

        params_tcc = current_params

        match optimizer:
            case dopyqo.VQEOptimizers.ExcitationSolve:
                counts_tcc = counts_tcc_excsolve
                values_tcc = values_tcc_excsolve
                tcc_vqe_params_lst = tcc_vqe_params_lst_excsolve
        self.tcc_vqe_params = params_tcc
        self.tcc_vqe_params_lst = tcc_vqe_params_lst

        self.tcc_vqe_counts = np.array(counts_tcc)
        self.tcc_vqe_values = np.array(values_tcc)

        self.tcc_vqe_result = res_tcc
        self.tcc_ansatz = ansatz

        self.vqe_counts = self.tcc_vqe_counts
        self.vqe_values = self.tcc_vqe_values
        self.vqe_result = self.tcc_vqe_result

        return self.tcc_vqe_result

    def solve_hf(self) -> float:
        """Perform a Hartree-Fock (HF) calculation with PySCF.
        The PySCF RHF object is saved in self.mf.

        Returns:
            float: HF energy
        """
        # https://github.com/pyscf/pyscf/blob/master/examples/scf/40-customizing_hamiltonian.py

        energy_offset = self.constants
        if isinstance(self.constants, dict):
            energy_offset = sum(self.constant.values())

        one_mo = self.h_pq
        mol = pyscf.gto.M()
        nelec = 0
        nelec = (self.nelec, self.nelec)
        nelec = sum(nelec)
        mol.nelectron = nelec
        mf = scf.RHF(mol)
        # mf.verbose = 5

        mf.get_hcore = lambda *args: one_mo.real
        mf.get_ovlp = lambda *args: np.eye(self.norb)
        mf._eri = pyscf.ao2mo.restore(8, self.h_pqrs.swapaxes(1, 2).swapaxes(1, 3).real, self.norb)
        mf.energy_nuc = lambda *args: energy_offset

        mol.incore_anyway = True
        res = mf.kernel()
        self.mo_coeff = mf.mo_coeff
        self.mf = mf
        return res

    def solve_fci(self, n_energies=1) -> np.ndarray:
        """Perform a FCI calculation using PySCF. If the matrix elements self.h_pq and self.h_pqrs are real,
        the function self.solve_fci_spin1 is executed which uses the PySCF solver pyscf.fci.direct_spin1.FCISolver().
        If the matrix elements are complex, the function self.solve_fci_general is executed which uses the PySCF solver pyscf.fci.direct_spin1.FCISolver().

        Args:
            n_energies (int, optional): Number of energies the FCI solver will compute. Defaults to 1.

        Returns:
            np.ndarray: Computed FCI energies as a numpy array.
        """
        energy_offset = self.constants
        if isinstance(self.constants, dict):
            energy_offset = sum(self.constant.values())

        if not np.allclose(self.h_pq.real, self.h_pq) or not np.allclose(self.h_pqrs.real, self.h_pqrs):
            return self.solve_fci_general(n_energies=n_energies)
        return self.solve_fci_spin1(n_energies=n_energies)

    def solve_fci_spin1(self, n_energies=1) -> np.ndarray:
        """Perform a FCI calculation using PySCF solver pyscf.fci.direct_spin1.FCISolver().

        Args:
            n_energies (int, optional): Number of energies the FCI solver will compute. Defaults to 1.

        Returns:
            np.ndarray: Computed FCI energies as a numpy array.
        """
        energy_offset = self.constants
        if isinstance(self.constants, dict):
            energy_offset = sum(self.constant.values())

        one_mo = self.h_pq
        two_mo = self.h_pqrs.swapaxes(1, 2).swapaxes(1, 3)
        if not np.allclose(one_mo.real, one_mo):
            print(
                f"{RED}Hamiltonian error (solve_fci_spin1): Single electron matrix elements are complex. Use solve_fci_general method instead!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        if not np.allclose(two_mo.real, two_mo):
            print(
                f"{RED}Hamiltonian error (solve_fci_spin1): Two electron matrix elements are complex. Use solve_fci_general method instead!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        if not np.allclose(energy_offset.real, energy_offset):
            print(
                f"{RED}Hamiltonian error (solve_fci_spin1): Constant energy offset is complex. Use solve_fci_general method instead!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)

        if self.h_pq.shape[0] == 0:
            return energy_offset

        nelec = (self.nelec, self.nelec)

        # FCI calculation
        nroots = n_energies  # number of states to calculate
        self.fci_solver = pyscf.fci.direct_spin1.FCISolver()
        h_pq_real = self.h_pq.real
        h_pqrs_real = self.h_pqrs.real.swapaxes(1, 2).swapaxes(1, 3)

        self.fci_evs, self.fci_evcs = self.fci_solver.kernel(
            h1e=h_pq_real,
            eri=h_pqrs_real,
            norb=self.norb,
            nelec=nelec,
            nroots=nroots,
        )
        # Full Hamiltonian matrix
        # h_fci = pyscf.fci.direct_spin1.pspace(h_pq_real, h_pqrs_real, self.norb, nelec, np=100000000)[1]

        # Save eigenvalues and -vectors in lists
        if n_energies == 1:
            self.fci_evs = np.array([self.fci_evs])
            self.fci_evcs = [self.fci_evcs]

        fci_energy = self.fci_evs + energy_offset

        self.fci_energy = fci_energy
        return fci_energy

    def solve_fci_uhf(self, n_energies=1) -> np.ndarray:
        """Perform a FCI calculation using the PySCF solver pyscf.fci.direct_uhf.FCISolver().

        Args:
            n_energies (int, optional): Number of energies the FCI solver will compute. Defaults to 1.

        Returns:
            np.ndarray: Computed FCI energies as a numpy array.
        """
        energy_offset = self.constants
        if isinstance(self.constants, dict):
            energy_offset = sum(self.constant.values())

        one_mo = self.h_pq
        two_mo = self.h_pqrs.swapaxes(1, 2).swapaxes(1, 3)
        if not np.allclose(one_mo.real, one_mo):
            print(
                f"{RED}Hamiltonian error (solve_fci_uhf): Single electron matrix elements are complex. Use solve_fci_general method instead!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        if not np.allclose(two_mo.real, two_mo):
            print(
                f"{RED}Hamiltonian error (solve_fci_uhf): Two electron matrix elements are complex. Use solve_fci_general method instead!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        if not np.allclose(energy_offset.real, energy_offset):
            print(
                f"{RED}Hamiltonian error (solve_fci_uhf): Constant energy offset is complex. Use solve_fci_general method instead!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)

        # Calculate matrix elements
        h_ij_up, h_ij_dw = self.h_pq, self.h_pq
        eri_up, eri_dw, eri_dw_up, eri_up_dw = (
            self.h_pqrs,
            self.h_pqrs,
            self.h_pqrs,
            self.h_pqrs,
        )

        # Transform ERIs to chemist's index order
        eri_up = eri_up.swapaxes(1, 2).swapaxes(1, 3)
        eri_dw = eri_dw.swapaxes(1, 2).swapaxes(1, 3)
        eri_dw_up = eri_dw_up.swapaxes(1, 2).swapaxes(1, 3)
        eri_up_dw = eri_up_dw.swapaxes(1, 2).swapaxes(1, 3)

        # TODO: Not optimal to calculate nelec manually. Add this to Wfc class?
        nelec = (self.nelec, self.nelec)
        # FCI calculation
        nroots = n_energies  # number of states to calculate
        norb = self.norb

        h_ij_up = h_ij_up.real
        h_ij_dw = h_ij_dw.real
        eri_up = eri_up.real
        eri_dw = eri_dw.real
        eri_up_dw = eri_up_dw.real

        self.fcisolver = pyscf.fci.direct_uhf.FCISolver()
        # Ordering of parameters from direct_uhf.make_hdiag
        self.fci_evs, self.fci_evcs = self.fcisolver.kernel(
            h1e=(h_ij_up, h_ij_dw),  # a, b (a=up, b=down)
            eri=(
                eri_up,
                eri_up_dw,
                eri_dw,
            ),  # aa, ab, bb (a=up, b=down)
            norb=norb,
            nelec=nelec,
            nroots=nroots,
        )

        # h_ij_up = h_ij_up.real
        # h_ij_up_diag = np.diag(h_ij_up.diagonal())
        # h_ij_up = np.zeros((4, 4))
        # h_ij_up[0, 0] = h_ij_up_diag[0, 0]
        # h_ij_up[1, 1] = h_ij_up_diag[1, 1]
        # h_ij_up[2, 2] = h_ij_up_diag[0, 0]
        # h_ij_up[3, 3] = h_ij_up_diag[1, 1]
        # eri_up = np.zeros((4, 4, 4, 4))
        # self.fcisolver = pyscf.fci.fci_dhf_slow.FCI()
        # self.fci_evs, self.fci_evcs = pyscf.fci.fci_dhf_slow.kernel(
        #     h1e=h_ij_up,
        #     eri=eri_up,
        #     norb=norb,
        #     nelec=sum(nelec),
        #     nroots=nroots,
        # )

        # self.fci_evs, self.fci_evcs = pyscf.fci.fci_slow.kernel(
        #     h1e=h_ij_up,
        #     eri=eri_up,
        #     norb=norb,
        #     nelec=sum(nelec),
        # )

        # Save eigenvalues and -vectors in lists
        if n_energies == 1:
            self.fci_evs = np.array([self.fci_evs])
            self.fci_evcs = [self.fci_evcs]

        fci_energy = self.fci_evs + energy_offset

        self.fci_energy = fci_energy
        return fci_energy

    def solve_fci_general(self, n_energies=1) -> np.ndarray:
        """Perform a FCI calculation using the PySCF solver pyscf.fci.direct_spin1.FCISolver().

        Args:
            n_energies (int, optional): Number of energies the FCI solver will compute. Defaults to 1.

        Returns:
            np.ndarray: Computed FCI energies as a numpy array.
        """
        # Calculate matrix elements
        h_ij_up, h_ij_dw = self.h_pq, self.h_pq
        eri_up, eri_dw, eri_dw_up, eri_up_dw = (
            self.h_pqrs,
            self.h_pqrs,
            self.h_pqrs,
            self.h_pqrs,
        )
        # Transform ERIs to chemist's index order
        eri_up = eri_up.swapaxes(1, 2).swapaxes(1, 3)
        eri_dw = eri_dw.swapaxes(1, 2).swapaxes(1, 3)
        eri_dw_up = eri_dw_up.swapaxes(1, 2).swapaxes(1, 3)
        eri_up_dw = eri_up_dw.swapaxes(1, 2).swapaxes(1, 3)

        # TODO: Not optimal to calculate nelec manually. Add this to Wfc class?
        nelec = (self.nelec, self.nelec)
        # FCI calculation
        nroots = n_energies  # number of states to calculate
        norb = self.norb

        energy_offset = self.constants
        if isinstance(self.constants, dict):
            energy_offset = sum(self.constant.values())

        # # Transforming matrices into fully spin-mixed form used in pyscf.scf.GHF and pyscf.fci.fci_dhf_slow
        # # We have separate matrices for up- and down-spin,
        # #       shape (norb, norb) for h1e and (norb, norb, norb, norb) for eri each
        # # We want one matrix for both spin-channels allowing also for spin-mixing,
        # #       shape (2*norb, 2*norb) for h1e and (2*norb, 2*norb, 2*norb, 2*norb) for eri each.
        # #       We set all spin-mixing matrix elements to zero, e.g. <\psi_{1,up}|h|\psi_{1,down}> = 0.0
        # # Orbital order is (α (up) and β (down) spin-channels):
        # #       β1, α1, β2, α2, β3, α3, β4, α4, β5, α5, .... where αn is the n-orbital with spin α
        # id_up = np.array([[1.0, 0.0], [0.0, 0.0]])
        # id_dw = np.array([[0.0, 0.0], [0.0, 1.0]])
        # h1e_custom_up = np.kron(h_ij_up, id_up)
        # h1e_custom_dw = np.kron(h_ij_dw, id_dw)
        # h1e_custom_spin = h1e_custom_up + h1e_custom_dw
        # eri_custom_up = np.kron(np.kron(eri_up, id_up).T, id_up).T
        # eri_custom_dw = np.kron(np.kron(eri_dw, id_dw).T, id_dw).T
        # eri_custom_dw_up = np.kron(np.kron(eri_dw_up, id_dw).T, id_up).T
        # eri_custom_up_dw = np.kron(np.kron(eri_up_dw, id_up).T, id_dw).T
        # eri_custom_spin = eri_custom_up + eri_custom_dw + eri_custom_dw_up + eri_custom_up_dw

        h1e_custom_spin = np.zeros((2 * norb, 2 * norb), dtype=np.complex128)
        for i in range(norb):
            for j in range(norb):
                h1e_custom_spin[2 * i, 2 * j] = h_ij_up[i, j]  # spin-↑
                h1e_custom_spin[2 * i + 1, 2 * j + 1] = h_ij_dw[i, j]  # spin-↓
        eri_custom_spin = np.zeros((2 * norb, 2 * norb, 2 * norb, 2 * norb), dtype=np.complex128)
        # (i,l) have same spin and (j,k) have same spin
        for i in range(norb):
            for j in range(norb):
                for k in range(norb):
                    for l in range(norb):
                        # += same as =
                        eri_custom_spin[2 * i, 2 * j, 2 * k, 2 * l] = self.h_pqrs[i, j, k, l]  # ↑↑
                        eri_custom_spin[2 * i + 1, 2 * j + 1, 2 * k + 1, 2 * l + 1] = self.h_pqrs[i, j, k, l]  # ↓↓
                        eri_custom_spin[2 * i, 2 * j + 1, 2 * k + 1, 2 * l] = self.h_pqrs[i, j, k, l]  # ↑↓
                        eri_custom_spin[2 * i + 1, 2 * j, 2 * k, 2 * l + 1] = self.h_pqrs[i, j, k, l]  # ↓↑
        eri_custom_spin = eri_custom_spin.transpose(0, 3, 1, 2)  #  / (-2.0)  # physicist to chemist order

        # self.fci_evs, self.fci_evcs = pyscf.fci.fci_dhf_slow.kernel(
        #     h1e=h1e_custom_spin,
        #     eri=eri_custom_spin,
        #     norb=norb * 2,
        #     nelec=sum(nelec),
        #     nroots=nroots,
        # )

        # self.fcisolver = pyscf.fci.fci_dhf_slow.FCI()
        # self.fci_evs, self.fci_evcs = self.fcisolver.kernel(
        #     h1e=h1e_custom_spin.real,  # TODO: REMOVE .real!
        #     eri=eri_custom_spin.real,
        #     norb=norb * 2,
        #     nelec=sum(nelec),
        #     nroots=nroots,
        #     # verbose=5,
        # )

        self.fcisolver = pyscf.fci.fci_dhf_slow.FCI()
        na = pyscf.fci.cistring.num_strings(norb * 2, sum(nelec))
        # random initial guess for robust results, see https://github.com/pyscf/pyscf/issues/2827
        # ci0 = [np.random.random((na,)) + np.random.random((na,)) * 1j]
        # self.fci_evs, self.fci_evcs = self.fcisolver.kernel(h1e_custom_spin, eri_custom_spin, norb * 2, nelec=sum(nelec), ci0=ci0)

        # TODO: Use following only if using updated pyscf.fci.fci_dhf_slow.py from
        #       https://github.com/pyscf/pyscf/commit/623f6f0a1d94972ead2d432d1e76e0590c021bbe
        self.fci_evs, self.fci_evcs = self.fcisolver.kernel(h1e_custom_spin, eri_custom_spin, norb * 2, nelec=sum(nelec))

        # Save eigenvalues and -vectors in lists
        if n_energies == 1:
            self.fci_evs = np.array([self.fci_evs])
            self.fci_evcs = [self.fci_evcs]

        fci_energy = self.fci_evs + energy_offset

        self.fci_energy = fci_energy
        return fci_energy

    def solve_CCSD(self) -> float:
        """Perform a CCSD calculation using the PySCF solver pyscf.cc.RCCSD().

        Returns:
            float:  Computed CCSD energy.
        """
        # based on https://github.com/pyscf/pyscf/blob/master/examples/cc/40-ccsd_custom_hamiltonian.py
        # Returns the correlation energy. To obtain the total energy HF energy needs to be added.

        # Construction of a molecule container. Only used as a means to
        # hand over the hamiltonian
        mol = gto.M(verbose=0)
        n = self.nelec * 2
        mol.nelectron = n
        mol.incore_anyway = True

        h1 = self.h_pq.real

        mf = scf.RHF(mol)
        mf.get_hcore = lambda *args: h1
        mf.get_ovlp = lambda *args: np.eye(h1.shape[0])

        eri = self.h_pqrs.real
        # Transform ERIs to chemist's index order
        eri = eri.swapaxes(1, 2).swapaxes(1, 3)
        mf._eri = eri

        mf.kernel()

        # pyscf constructs molecular orbitals from atomic orbitals and stores the coefficients
        # for that in mo_coeff. As we already hand over matrices that are calculated with molecular orbitals and
        # we do not want to hand over thousands of AOs, setting mo_coeff = eye() leads to AO=MO
        mf.mo_coeff = np.eye(mf.mo_coeff.shape[0])

        mycc = cc.RCCSD(mf)
        mycc.out = mycc.kernel()

        return mycc.e_corr + self.hf_energy()

    def solve_CCSD_t(self) -> np.ndarray:
        """Perform a CCSD(T) calculation using the PySCF solver pyscf.cc.CCSD() in combination with the .ccsd_t() PySCF function.

        Returns:
            float:  Computed CCSD(T) energy.
        """
        mol = gto.M(verbose=0)
        n = self.nelec * 2
        mol.nelectron = n
        mol.incore_anyway = True

        h1 = self.h_pq.real

        mf = scf.RHF(mol)
        mf.get_hcore = lambda *args: h1
        mf.get_ovlp = lambda *args: np.eye(h1.shape[0])

        eri = self.h_pqrs.real
        eri = eri.swapaxes(1, 2).swapaxes(1, 3)
        mf._eri = eri

        mf.kernel()

        mf.mo_coeff = np.eye(mf.mo_coeff.shape[0])

        # 1. Normal CCSD calculation
        mycc = cc.CCSD(mf).run()
        mycc.kernel()
        # 2. Correction of the correlation energy due to triples
        et = mycc.ccsd_t()

        return mycc.e_corr + et + self.hf_energy()



from enum import Enum


class ExcitationPools(Enum):
    SINGLES = 1
    DOUBLES = 2
    SINGLES_DOUBLES = 3


class VQEOptimizers(Enum):
    L_BFGS_B = 1
    COBYLA = 2
    ExcitationSolve = 3


class AdaptSelectionCriterion(Enum):
    """Types of selection criterions used in ADAPT-VQE"""

    GRADIENT = 1
    """Gradient-based selection as proposed by Grimsley et al. 2018 (https://doi.org/10.1038/s41467-019-10988-2),
    by estimating the operator impact by calculating its energy gradient at parameter value 0.
    """
    ENERGY = 2
    """Energy-based selection as proposed by Jäger et al. 2025 (https://doi.org/10.48550/arXiv.2409.05939),
    by estimating the operator impact by calculating its maximum energy impact at its optimal parameter value.
    """



import re
from collections.abc import Iterable
import dopyqo
from dopyqo.colors import *


def active_space_string(
    occupations: Iterable[float],
    indices_active_space: Iterable[int],
    occs_per_line: int = 8,
    color_active_space: str = SOFT_RED,
    color_rest: str = MEDIUM_GRAY,
    separator: str = " ",
) -> str:
    """Generate string of occupations where an active space is marked

    Args:
        occupations (Iterable[float]): Iterable holding the occupations
        indices_active_space (Iterable[int]): Iterable holding the indices of the active space
        occ_per_line (int): Number of occupations listed in one line in the output string. 8 is used in a QE output file. Defaults to 8.
        color_active_space (str): Color of occupations in the active space
        color_rest (str): Color of all occupations not in the active space
        separator (str): Separator used to enclose the active space
    Returns:
        str: String of occupations where the active space is marked
    """
    occ_str = " ".join(
        [
            (f"|{val}" if i == indices_active_space[0] else f"{val}|" if i == indices_active_space[-1] else f"{val}")
            for i, val in enumerate(occupations)
        ]
    )
    occ_str = re.sub(r"\s*\|\s*", "|", occ_str)  # remove spaces around every |
    occ_str_split = re.findall(r"\|\d|\d", occ_str)  # create list by splitting at " ", and make "a|b" into "a" and "|b"
    # Now split string into lines containing group_size number of elements
    # Create sublists containing group_size number of elements
    occ_str_grouped = [occ_str_split[i : i + occs_per_line] for i in range(0, len(occ_str_split), occs_per_line)]
    # Join sublists with empty spaces
    occ_str_grouped = [(" " * len(separator)).join(group) for group in occ_str_grouped]
    # Remove spaces around every |
    occ_str_grouped = [re.sub(r"\s*\|\s*", "|", x) for x in occ_str_grouped]
    # Indent every sublist/line by an empty space, do not indent if line starts with first |, if line starts with second | append this | to line before
    # We indent so that we can place the first | at the start of the line if line starts with the first |
    occ_str_split = []
    act_spc_start_found = False
    for line in occ_str_grouped:
        if line.startswith("|") and not act_spc_start_found:  # the | here signals the start of the active space
            occ_str_split.append(line)
        elif line.startswith("|") and act_spc_start_found:  # the | here signals the end of the active space
            occ_str_split[-1] = occ_str_split[-1] + "|"  # add | to end of previous line
            occ_str_split.append(len(separator) * " " + line[1:])  # add indented line without starting |
        else:
            occ_str_split.append(len(separator) * " " + line)  # add indented line
        if "|" in line and not act_spc_start_found:  # | in line but not at the start
            act_spc_start_found = True
    # Add color to active space and rest
    occ_str_split_w_color = []
    first_pipe_found = False
    for line in occ_str_split:  # Add color to active space
        parts = line.split("|")  # there can be at most two | in one line, yielding 3 parts
        if len(parts) == 3:  # both | are in the line
            # line = parts[0] + f"{color_active_space}|" + parts[1] + f"|{color_rest}" + parts[2] # with |
            # line = parts[0] + f"{color_active_space} " + parts[1] + f" {color_rest}" + parts[2]  # without |
            line = parts[0] + f"{color_active_space}{separator}" + parts[1] + f"{separator}{color_rest}" + parts[2]
        elif len(parts) == 2 and not first_pipe_found:  # first | is in the line
            first_pipe_found = True
            # line = parts[0] + f"{color_active_space}|" + parts[1] # with |
            # line = parts[0] + f"{SOFT_RED} " + parts[1]  # without |
            line = parts[0] + f"{color_active_space}{separator}" + parts[1]
        elif len(parts) == 2 and first_pipe_found:  # second | is in the line
            # line = parts[0] + f"|{color_rest}" + parts[1] # with |
            # line = parts[0] + f" {color_rest}" + parts[1]  # without |
            line = parts[0] + f"{separator}{color_rest}" + parts[1]
        else:  # no | is in the line
            line = line
        occ_str_split_w_color.append(line)
    occ_str = color_rest + "\n".join(occ_str_split_w_color) + RESET_COLOR

    return occ_str


def print_block(text: str, width: int = 60, color: str = NO_COLOR, flush: bool = False):
    text_max_len = max([len(x) for x in text.split("\n")])
    text_lines = len(text.split("\n"))
    if width < text_max_len + 2:
        width = text_max_len + 2
    prefix = color
    postfix = RESET_COLOR if color != NO_COLOR else ""
    print(prefix + "\t+" + "-" * width + "+")
    for line in text.split("\n"):
        print("\t|" + (width - len(line)) // 2 * " " + line + (width - len(line) - (width - len(line)) // 2) * " " + "|")
    print("\t+" + "-" * width + "+" + postfix, flush=flush)


def print_banner(banner, flush: bool = False):
    second_title = f"\nMany-body analysis on top of Quantum ESPRESSO calculations\nVersion: {dopyqo.__version__}"
    print_block("\n".join([banner.strip("\n"), second_title]), color=BRIGHT_CYAN, flush=flush)



import os
import numpy as np
import matplotlib.pyplot as plt
import dopyqo
from dopyqo.colors import *

if __name__ == "__main__":
    vqe_optimizer = dopyqo.VQEOptimizers.L_BFGS_B
    config = dopyqo.DopyqoConfig(
        base_folder=os.path.join("qe_files", "Be"),
        prefix="Be",
        active_electrons=2,
        active_orbitals=4,
        run_fci=True,
        run_vqe=True,
        vqe_adapt=True,
        vqe_adapt_drain_pool=True,
        vqe_optimizer=vqe_optimizer,
        vqe_excitations=dopyqo.ExcitationPools.SINGLES_DOUBLES,
        n_threads=10,
    )
    energy_dict, wfc_obj, h_ks, mats = dopyqo.run(config)
    dft_energy = energy_dict["dft_energy"]
    fci_energy = energy_dict["fci_energy"]
    vqe_energy = energy_dict["vqe_energy"]

    plt.plot(h_ks.vqe_counts, np.abs(h_ks.vqe_values - fci_energy), linestyle="-", marker="x")
    plt.yscale("log")
    plt.grid()



import sys
import itertools
import numpy as np
from itertools import product
import pyvista as pv
from dopyqo.colors import *

# http://www.xcrysden.org/doc/XSF.html
# Column-major order


def load_xsf(filename):
    with open(filename) as f:
        lines = f.readlines()

    data = None
    grid_vecs = None
    origin = None
    atom_pos = None
    primvecs = None
    convvecs = None
    for i, line in enumerate(lines):
        if line.strip().startswith("#"):
            continue
        if line.strip().startswith("PRIMVEC"):
            str_tmp = " ".join([x.strip() for x in lines[i + 1 : i + 4]]).strip()
            primvecs = np.fromstring(str_tmp, dtype=float, sep=" ").reshape((3, 3))
            # print(f"{primvecs=}")
        if line.strip().startswith("CONVVEC"):
            str_tmp = " ".join([x.strip() for x in lines[i + 1 : i + 4]]).strip()
            convvecs = np.fromstring(str_tmp, dtype=float, sep=" ").reshape((3, 3))
            # print(f"{convvecs=}")
        if line.strip().startswith("PRIMCOORD"):
            natm = int(lines[i + 1].strip().split()[0])
            # print(f"{natm=}")
            str_tmp = [x.strip() for x in lines[i + 2 : i + natm + 2]]
            str_tmp = " ".join([" ".join(x.split()[1:]) for x in str_tmp])  # 0 element is element string, e.g. Mg
            atom_pos = np.fromstring(str_tmp, dtype=float, sep=" ").reshape((natm, 3))  # In Angstrom
            # print(f"{atom_pos=}")
        if line.strip().startswith("BEGIN_DATAGRID"):
            nx, ny, nz = map(int, lines[i + 1].split())
            # print(f"{nx=} {ny=} {nz=}")
            origin = np.fromstring(lines[i + 2].strip(), dtype=float, sep=" ")
            # print(f"{origin=}")
            str_tmp = " ".join([x.strip() for x in lines[i + 3 : i + 6]]).strip()
            grid_vecs = np.fromstring(str_tmp, dtype=float, sep=" ").reshape((3, 3))
            # print(f"{grid_vecs=}")

            for i_tmp, line_tmp in enumerate(lines[i + 6 :]):
                if line_tmp.strip().startswith("END"):
                    break

            data_lst = []
            for data_line in lines[i + 6 : i + 6 + i_tmp]:
                data_lst.extend(map(float, data_line.strip().split()))
            # print(f"{len(data_lst)=}")
            # print(f"{nx * ny * nz =}")
            assert len(data_lst) == nx * ny * nz, f"Number of data points ({len(data_lst)}) is not equal to number of grid points ({nx*ny*nz})"

            data = np.array(data_lst).reshape((nx, ny, nz), order="C")
            # NOTE: Order C is needed to get correct plots for H2 compared to plots with VESTA.
            #       But for diamond order F looked fine.

    if data is None:
        print(f"{RED}XSF error: Could not read data from xsf file {filename}.{RESET_COLOR}")
        sys.exit(1)
    if grid_vecs is None:
        print(f"{RED}XSF error: Could not read vectors spanning the data grid from xsf file {filename}.{RESET_COLOR}")
        sys.exit(1)
    if origin is None:
        print(f"{RED}XSF error: Could not read origin of the data grid from xsf file {filename}.{RESET_COLOR}")
        sys.exit(1)
    if atom_pos is None:
        print(f"{RED}XSF error: Could not read atom positions from xsf file {filename}.{RESET_COLOR}")
        sys.exit(1)
    if primvecs is None:
        print(f"{RED}XSF error: Could not read primitive lattice vectors from xsf file {filename}.{RESET_COLOR}")
        sys.exit(1)
    if convvecs is None:
        print(
            f"{ORANGE}XSF warning: Could not read conventional lattice vectors from xsf file {filename}. "
            + f"Setting conventional lattice vectors equal to primitive lattice vectors.{RESET_COLOR}"
        )
        convvecs = primvecs.copy()

    return data, grid_vecs, origin, atom_pos, primvecs, convvecs


def lattice_points(a, b, c, counts, origin=None):
    a = np.array(a)
    b = np.array(b)
    c = np.array(c)
    n_a, n_b, n_c = counts
    if origin is None:
        origin = np.zeros(3)
    else:
        origin = np.array(origin)

    coords = []
    for i, j, k in product(range(n_a), range(n_b), range(n_c)):
        pt = origin + i * a + j * b + k * c
        coords.append(pt)
    return np.array(coords).reshape(n_a, n_b, n_c, 3)



import os
import numpy as np

if __name__ == "__main__":
    base_folder = os.path.join("qe_files", "Be")
    prefix = "Be"

    active_electrons = 2
    active_orbitals = 2

    data = np.load(os.path.join(base_folder, f"energies_{active_electrons}e_{active_orbitals}o.npz"), allow_pickle=True)
    for key, val in data.items():
        print(f"{key}: {val}")



import sys
import time
import logging
from math import prod
from contextlib import nullcontext
import numpy as np
from numba import njit
from dopyqo import calc_matrix_elements
from dopyqo.pseudopot import Pseudopot
from dopyqo import calc_pseudo_pot
from dopyqo.colors import *

# Calculate ERIs via h_ijkl = 4\pi \sum_p (\rho*_il(p) \rho_jk(p))/p²
# with \rho_ij(p)=\int dr \rho_ij(r) e^(-ipr) which is the Fourier transform of
# \rho_ij(r) where \rho_ij(r)=\psi*_i(r)\psi_j(r). Therefore \rho_ij(p)
# is the convolution between \psi*_i(p) and \psi_j(p): \psi*_i(p) * \psi_j(p)
# where \psi_i(p) is the Fourier transform of \psi_i(r)

# The calculation steps are inspired by WEST: https://west-code.org/, https://github.com/west-code-development/West
# Especially the code in the compute_eri_vc function: https://github.com/west-code-development/West/blob/master/Wfreq/solve_eri.f90#L327
# Publication related to WEST:
# Large Scale GW Calculations, M. Govoni and G. Galli, J. Chem. Theory Comput. 11, 2680 (2015)
# GPU Acceleration of Large-Scale Full-Frequency GW Calculations, V. Yu and M. Govoni, J. Chem. Theory Comput. 18, 4690 (2022)


@njit
def make_one_over_p_norm_squared_array(
    one_over_p_norm_squared_array: np.ndarray,
    b: np.ndarray,
    fft_grid: np.ndarray,
):
    """Calculate |p|² for each grid point
    Its important to calculate 1/|p|² for the whole momentum grid not only for
    momentum present in the used Miller indices, because
    the density \rho_ij(p) may extend further than the Miller indices,
    since \rho_ij(p) is the result of a convolution between two p-space wavefunctions
    """
    nx, ny, nz = one_over_p_norm_squared_array.shape
    half_nx = fft_grid[0] // 2
    half_ny = fft_grid[1] // 2
    half_nz = fft_grid[2] // 2

    for i in range(nx):
        x = i - half_nx
        px = b[0] * x
        for j in range(ny):
            y = j - half_ny
            py = b[1] * y
            for k in range(nz):
                z = k - half_nz
                p_norm = np.linalg.norm(px + py + b[2] * z, ord=2)
                p_norm_2 = p_norm**2
                if p_norm_2 < 1e-8:
                    one_over_p_norm_squared_array[i, j, k] = 0.0
                    continue

                one_over_p_norm_squared_array[i, j, k] = 1.0 / p_norm_2

    return one_over_p_norm_squared_array


def pair_density(
    c_ip_array: np.ndarray,
    c_jp_array: np.ndarray | None = None,
    use_gpu: bool = True,
) -> np.ndarray:
    r"""Calculate pair density in reciprocal space via Fourier transforms
    Calculates \rho_ij(p)=\psi*_i(p) * \psi_j(p) (* is convolution) which is the
    Fourier transform of \psi*_i(r) * \psi_j(r) (* is standard multiplication)

    The argument c_ip_array is \psi_i(p) for a p-grid in reciprocal space.
    The inverse Fourier transform of \psi_i(p) (or c_ip_array) is \psi_i(r) in real space.
    To calculate \psi*_i(r) * \psi_j(r) (* is standard multiplication) we inverse Fourier
    transform both \psi_i(p) and \psi_j(p) resulting in \psi_i(r) and \psi_j(r)
    and then perform the standard multiplication \psi*_i(r) * \psi_j(r) (* is standard multiplication).
    We then perform a Fourier transform of the calculated real space pair density
    \psi*_i(r) * \psi_j(r) (* is standard multiplication) and return the result.

    Args:
        c_ip_array (np.ndarray): Array of coefficients describing the Kohn-Sham orbitals \psi_i(p) in the plane wave basis,
                                 shape (#bands, #grid_size, #grid_size, #grid_size)
        c_jp_array (np.ndarray | None): Same as c_ip_array but the coefficients describing \psi_j(p)
                                        instead of \psi_i(p) in \rho_ij(p)=\psi*_i(p) * \psi_j(p) (* is convolution).
                                        If None, this is set to c_ip_array

    Returns:
        np.ndarray: Pair density in reciprocal space
    """
    using_cupy = False

    if use_gpu:
        try:
            import cupy as cp
        except ImportError:
            print(f"{ORANGE}Import warning: Could not import cupy package. Falling back to numpy.{RESET_COLOR}")
            xp = np
        else:  # No exception
            xp = cp
            using_cupy = True
    else:
        xp = np

    my_context = nullcontext()
    if using_cupy:
        my_context = cp.cuda.Device(cp.cuda.runtime.getDeviceCount() - 1)  # Use last available GPU
    with my_context:
        c_jp_array_not_given = False
        if c_jp_array is None:
            c_jp_array_not_given = True
            # c_jp_array = c_ip_array.copy()
        # print(f"{c_jp_array_not_given=}")

        dtype = xp.complex128

        if not c_jp_array_not_given:
            assert c_ip_array.ndim == c_jp_array.ndim == 4
            assert c_ip_array.shape[1:] == c_jp_array.shape[1:]

        nbands_i = c_ip_array.shape[0]
        nbands_j = nbands_i
        if not c_jp_array_not_given:
            nbands_j = c_jp_array.shape[0]
        ngrid = c_ip_array.shape[1:]
        if not c_jp_array_not_given:
            assert ngrid == c_jp_array.shape[1:]

        # rho_ij_p = xp.zeros((nbands_i, nbands_j, *ngrid), dtype)
        rho_ij_p = np.zeros((nbands_i, nbands_j, *ngrid), dtype)

        for i in range(nbands_i):
            logging.info("Calculating pair density %s/%s ...", i + 1, nbands_i)
            c_ip_shifted = xp.fft.ifftshift(c_ip_array[i])
            psi_r_i_conj = xp.fft.ifftn(c_ip_shifted).conj()
            #
            if c_jp_array_not_given:
                range_start = i
            else:
                range_start = 0
            #
            for j in range(range_start, nbands_j):
                if c_jp_array_not_given:
                    c_jp_shifted = xp.fft.ifftshift(c_ip_array[j])
                else:
                    c_jp_shifted = xp.fft.ifftshift(c_jp_array[j])
                psi_r_j = xp.fft.ifftn(c_jp_shifted)

                # psi*_i(r) . psi_j(r), where . is the standard multiplication
                # Same as psi_i(p) * psi_j(p), where * is the convolution operation
                rho_ij_r = psi_r_i_conj * psi_r_j

                rho_ij_p_val = xp.fft.fftshift(xp.fft.fftn(rho_ij_r))

                if using_cupy:
                    rho_ij_p_val = xp.asnumpy(rho_ij_p_val)
                rho_ij_p[i, j, :] = rho_ij_p_val
                if c_jp_array_not_given and i != j:
                    # \rho_ji(p)=\rho*_ij(-p)
                    rho_ij_p[j, i, :] = np.flip(rho_ij_p[i, j, :].conj(), (0, 1, 2))

    # if using_cupy:
    #     return xp.asnumpy(rho_ij_p)
    return rho_ij_p


def pair_density_real_space(
    c_ip_array: np.ndarray,
    c_jp_array: np.ndarray | None = None,
) -> np.ndarray:
    r"""Calculate pair density in reciprocal space via Fourier transforms
    Calculates \psi*_i(r) * \psi_j(r) (* is standard multiplication)

    The argument c_ip_array is \psi_i(p) for a p-grid in reciprocal space.
    The inverse Fourier transform of \psi_i(p) (or c_ip_array) is \psi_i(r) in real space.
    To calculate \psi*_i(r) * \psi_j(r) (* is standard multiplication) we inverse Fourier
    transform both \psi_i(p) and \psi_j(p) resulting in \psi_i(r) and \psi_j(r)
    and then perform the standard multiplication \psi*_i(r) * \psi_j(r) (* is standard multiplication).
    We return the calculated real space pair density.

    Args:
        c_ip_array (np.ndarray): Array of coefficients describing the Kohn-Sham orbitals \psi_i(p) in the plane wave basis,
                                 shape (#bands, #grid_size, #grid_size, #grid_size)
        c_jp_array (np.ndarray | None): Same as c_ip_array but the coefficients describing \psi_j(p)
                                        instead of \psi_i(p) in \rho_ij(p)=\psi*_i(p) * \psi_j(p) (* is convolution).
                                        If None, this is set to c_ip_array

    Returns:
        np.ndarray: Pair density in real space
    """
    single_coeffs = False
    if c_jp_array is None:
        single_coeffs = True
        c_jp_array = c_ip_array.copy()

    dtype = np.complex128

    assert c_ip_array.ndim == c_jp_array.ndim == 4
    if not single_coeffs:
        assert c_ip_array.shape[1:] == c_jp_array.shape[1:]

    nbands_i = c_ip_array.shape[0]
    nbands_j = nbands_i
    if not single_coeffs:
        nbands_j = c_jp_array.shape[0]
    ngrid = c_ip_array.shape[1:]
    if not single_coeffs:
        assert ngrid == c_jp_array.shape[1:]

    rho_ij_r_mat = np.zeros((nbands_i, nbands_j, *ngrid), dtype)

    for i in range(nbands_i):
        logging.info("Calculating pair density %s/%s ...", i + 1, nbands_i)
        c_ip_shifted = np.fft.ifftshift(c_ip_array[i])
        psi_r_i_conj = np.fft.ifftn(c_ip_shifted).conj()
        for j in range(nbands_j):
            c_jp_shifted = np.fft.ifftshift(c_jp_array[j])
            psi_r_j = np.fft.ifftn(c_jp_shifted)

            # psi*_i(r) . psi_j(r), where . is the standard multiplication
            # Same as psi_i(p) * psi_j(p), where * is the convolution operation
            rho_ij_r = psi_r_i_conj * psi_r_j

            rho_ij_r_mat[i, j, :] = rho_ij_r

    return rho_ij_r_mat


def eri(
    c_ip: np.ndarray,
    b: np.ndarray,
    mill: np.ndarray,
    fft_grid: np.ndarray | None = None,
    use_gpu: bool = True,
) -> np.ndarray:
    r"""Calculate Electron Repulsion Integrals (ERIs) via pair densities in the physicists' index order
    We calculate

    .. math::

        h_{ijkl}=4\pi \sum_{p \neq 0} \rho_{il}(-p)\rho_{jk}(p)/|p|²
                =4\pi \sum_{p \neq 0} \rho*_{li}(p)\rho_{jk}(p)/|p|²

    where ijkl are indexing spatial orbitals
    Since the momenta p and reciprocal space wavefunctions \psi_i(p) (given as c_ip)
    are given as a list, we need to transfer \psi_i(p) to a 3D grid.
    Then \psi_i(p) is represented on a 3D reciprocal space grid.
    We calculate 1/|p|^2 on all grid points resulting in an infinite value at the
    center of the grid. This infinite value is set to zero, therefore,
    technically we later perform a sum over all momenta p except the zero momenta.
    Note that there are different techniques handling with this singularity
    by e.g. using a resolution-of-identity.

    Note that h_{ijkl} does not explicitly depend on the k-point but only implicitly
    depends on the k-point via the plane wave coefficients c_ip.
    Any explicit dependecy is cancelled out since the k-point contributes a phase
    to each KS-orbital (e^(ikr)) but since the pair densities are a product of two
    orbitals, where one in complex conjugated and the other is not, the phases
    of the two orbitals cancel out ([e^(ikr)]* e^(ikr) = e^(-ikr) e^(ikr) = 1)

    Args:
        c_ip (np.ndarray): Array of coefficients describing the Kohn-Sham orbitals in the plane wave basis, shape (#bands, #waves)
        b (np.ndarray): Array of reciprocal lattice vectors, shape (3, 3)
        mill (np.ndarray): Array of Miller indices, shape (#waves, 3)
        fft_grid_size (np.ndarray | None): Edge lengths of the use FFT grid. Has to be larger or equal to grid
            where wavefunctions are defined on. Defaults to None, for which the wavefunction
            grid is used.

    Returns:
        np.ndarray: ERIs in reciprocal space
    """
    nbands, _nwaves = c_ip.shape  # number of DFT bands and number of plane waves

    assert mill.ndim == 2 and mill.shape[1] == 3, f"Array mill should contain 3D Miller indices but has shape {mill.shape}!"
    assert b.ndim == 2 and b.shape[0] == b.shape[1] == 3, f"Array b should contain 3 3D reciprocal lattice vectors but has shape {b.shape}!"

    max_min = mill.max(axis=0) - mill.min(axis=0)
    if fft_grid is not None:
        assert isinstance(fft_grid, np.ndarray), f"fft_grid ({fft_grid}, type {type(fft_grid)}) is not a numpy array!"
        assert fft_grid.shape == (3,), f"fft_grid (shape {fft_grid.shape}) has to have shape (3,), one grid length for each direction!"
        assert np.all(fft_grid >= 0), f"fft_grid {fft_grid} has values smaller than zero!"
        assert np.all(fft_grid >= max_min), f"fft_grid ({fft_grid}) needs to be larger or equal to wavefunction grid ({max_min})!"

        max_min = fft_grid.copy()

    # Grid needs to have odd edge length values so that the zero momentum is in the center of the grid
    for i, val in enumerate(max_min):
        if val % 2 == 0:
            max_min[i] += 1

    logging.info("FFT grid: %s", max_min)

    # Initialize 3D array for each DFT band with zero
    # We want to define \psi_i(p) as a matrix on the whole momentum grid for each i, instead of a list
    # *max_min is number of grid points for given maximum and minimum momenta and given grid spacing
    c_ip_array = np.zeros(
        (
            nbands,
            *max_min,
        ),
        dtype=c_ip.dtype,
    )

    # Set \psi_i(p) on given grid points
    assert mill.shape[0] == c_ip.shape[1]
    for idx, mill_idx in enumerate(mill):
        x, y, z = mill_idx
        i, j, k = (
            x + max_min[0] // 2,
            y + max_min[1] // 2,
            z + max_min[2] // 2,
        )
        c_ip_array[:, i, j, k] = c_ip[:, idx]

    start_time = time.perf_counter()
    one_over_p_norm_squared_array = np.zeros(max_min)
    one_over_p_norm_squared_array = make_one_over_p_norm_squared_array(one_over_p_norm_squared_array, b, max_min)
    norm_mat_time = time.perf_counter() - start_time

    # Calculate pair density \rho_ij(p) in reciprocal space
    start_time = time.perf_counter()
    rho_ij_p = pair_density(c_ip_array=c_ip_array, use_gpu=use_gpu)  # , c_jp_array=c_ip_array)
    pair_density_time = time.perf_counter() - start_time
    # Initialize ERI array
    # TODO: We do not need to calculate all matrix elements due to symmetries
    eri_mat = np.zeros((nbands, nbands, nbands, nbands), dtype=c_ip.dtype)

    start_time = time.perf_counter()
    # Calculate ERI matrix and rescale Fourier transforms
    # h_ijkl = 4\pi \sum_{p \neq 0} \rho*_li(p)\rho_jk(p)/|p|²
    # h_ijkl = 4\pi \sum_{p \neq 0} \rho_il(-p)\rho_jk(p)/|p|²
    eri_mat = (
        4
        * np.pi
        * np.einsum(
            "ilxyz, jkxyz, xyz -> ijkl",  # Physicists' order
            # * prod(max_min) for rescaling of Fourier transforms (order ~1e6)
            # np.flip(rho_ij_p, (2, 3, 4)) * prod(max_min),  # \rho_il(-p)
            rho_ij_p.conj().swapaxes(0, 1) * prod(max_min),  # \rho*_li(p) = \rho_il(-p)
            rho_ij_p * prod(max_min),
            one_over_p_norm_squared_array,
            optimize=True,  # larger memory footprint but faster
        )
    )
    einsum_time = time.perf_counter() - start_time
    # print(f"Time summary: 1/|p|^2: {norm_mat_time:.3f}s | pair density: {pair_density_time:.3f}s | einsum: {einsum_time:.3f}s")
    return eri_mat


def get_frozen_core_energy_eri(
    c_ip_core: np.ndarray,
    b: np.ndarray,
    mill: np.ndarray,
    cell_volume: float,
    fft_grid: np.ndarray | None = None,
    rho_ij_p: np.ndarray | None = None,
    one_over_p_norm_squared_array: np.ndarray | None = None,
    use_gpu: bool = True,
) -> float:
    r"""Calculate frozen core energy ERI part \sum_{ij}^{\mathrm{frozen}} (2h_{iijj} - h_{ijji})

    Args:
        p (np.ndarray): Array of momentum vectors, shape (#waves, 3)
        c_ip_core (np.ndarray): Array of coefficients describing the active Kohn-Sham orbitals in the plane wave basis, shape (#bands, #waves)
        b (np.ndarray): Array of reciprocal lattice vectors, shape (3, 3)
        mill (np.ndarray): Array of Miller indices, shape (#waves, 3)
        cell_volume (float): Cell volume of the computationen real space cell.
        atom_positions (np.ndarray): 1D array of positions of each atom
        atomic_numbers (np.ndarray): 1D array of atomic number of each atom
        occupations_core (np.ndarray): Array of occupations of the core orbitals.
        rho_ij_p (np.ndarray | None): Array of pair densities of shape (#bands, #bands, #nwaves_fft),
            where #nwaves_fft is the number of points in momentum space that are used in the Fourier transforms.
            Is calculated from c_ip_core if None. If not None one_over_p_norm_squared_array has to be given, too.
            Defaults to None.
        one_over_p_norm_squared_array (np.ndarray | None): Array of 1/|p|^2 of shape (#nwaves_fft),
            where #nwaves_fft is the same as for rho_ij_p. Is calculated from b and mill if None.
            If not None rho_ij_p has to be given, too. Defaults to None.

    Returns:
        np.ndarray: ERIs in reciprocal space
    """
    max_min = mill.max(axis=0) - mill.min(axis=0)
    if fft_grid is not None:
        assert isinstance(fft_grid, np.ndarray), f"fft_grid ({fft_grid}, type {type(fft_grid)}) is not a numpy array!"
        assert fft_grid.shape == (3,), f"fft_grid (shape {fft_grid.shape}) has to have shape (3,)"
        assert np.all(fft_grid >= 0), f"fft_grid {fft_grid} has values smaller than zero!"
        assert np.all(fft_grid >= max_min), f"fft_grid ({fft_grid}) needs to be larger or equalt than wavefunction grid ({max_min})!"

        max_min = fft_grid.copy()

    # Grid needs to have odd edge length values so that the zero momentum is in the center of the grid
    for i, val in enumerate(max_min):
        if val % 2 == 0:
            max_min[i] += 1

    logging.info("FFT grid: %s", max_min)

    pre_calc = False  # Are the pair densities and 1/p² values given, if not calculated them
    if rho_ij_p is not None or one_over_p_norm_squared_array is not None:
        assert (
            rho_ij_p is not None and one_over_p_norm_squared_array is not None
        ), "Both rho_ij_p and one_over_p_norm_squared_array have to be given if one is given, but only one is given!"
        assert rho_ij_p.shape[0] == rho_ij_p.shape[1] == rho_ij_p.shape[0] == c_ip_core.shape[0]
        assert one_over_p_norm_squared_array.shape[0] == rho_ij_p.shape[2]
        pre_calc = True

    if not pre_calc:
        nbands, _nwaves = c_ip_core.shape  # number of DFT bands and number of plane waves

        assert mill.ndim == 2 and mill.shape[1] == 3, f"Array mill should contain 3D Miller indices but has shape {mill.shape}!"
        assert b.ndim == 2 and b.shape[0] == b.shape[1] == 3, f"Array b should contain 3 3D reciprocal lattice vectors but has shape {b.shape}!"

        # Initialize 3D array for each DFT band with zero
        # We want to define \psi_i(p) as a matrix on the whole momentum grid for each i, instead of a list
        c_ip_array = np.zeros(
            (
                nbands,
                # Number of grid points for given maximum and minimum momenta and given grid spacing
                *max_min,
            ),
            dtype=c_ip_core.dtype,
        )
        # Set \psi_i(p) on given grid points
        for idx, mill_idx in enumerate(mill):
            x, y, z = mill_idx
            i, j, k = (
                x + max_min[0] // 2,
                y + max_min[1] // 2,
                z + max_min[2] // 2,
            )
            c_ip_array[:, i, j, k] = c_ip_core[:, idx]

        one_over_p_norm_squared_array = np.zeros(max_min)
        one_over_p_norm_squared_array = make_one_over_p_norm_squared_array(one_over_p_norm_squared_array, b, max_min)

        # Calculate pair density \rho_ij(p) in reciprocal space
        rho_ij_p = pair_density(c_ip_array=c_ip_array, use_gpu=use_gpu)

    logging.info("Calc einsums (in get_frozen_core_energy_eri)...")
    # 4\pi \sum_{ij} \sum_{p \neq 0} (2\rho_ij(-p)\rho_ij(p)-\rho_ii(-p)\rho_jj(p))/|p|²
    #   and we get \rho_il(-p) from \rho_il(p) via inverting the x,y,z axes.
    #   alternatively one could use \rho_il(-p)=\rho*_li(p), which cannot be easily performed
    #       if i and l belong to different spaces, e.g. i belongs to the core and l belongs
    #       to an active space. And np.flip(...) seems to be faster than .conj() here.
    # 2 g_iijj - g_ijji, like in Yalouz et al, but they use chemists order
    # We use physicist order, therefore we use 2 g_ijji - g_ijij
    rho_ij_p_prod = rho_ij_p * prod(max_min)
    rho_ij_p_flip_prod = np.flip(rho_ij_p_prod, (2, 3, 4))
    eri_energy = (
        4
        * np.pi
        * (
            2
            * np.einsum(
                "iixyz, jjxyz, xyz -> ij",  # ijji
                rho_ij_p_flip_prod,
                rho_ij_p_prod,
                one_over_p_norm_squared_array,
                optimize=True,  # larger memory footprint but faster
            ).sum()
            - np.einsum(
                "ijxyz, jixyz, xyz -> ij",  # ijij
                rho_ij_p_flip_prod,
                rho_ij_p_prod,
                one_over_p_norm_squared_array,
                optimize=True,  # larger memory footprint but faster
            ).sum()
        )
    )

    # Rescaling of Fourier transforms
    return eri_energy / cell_volume


def get_frozen_core_energy(
    p: np.ndarray,
    c_ip_core: np.ndarray,
    b: np.ndarray,
    mill: np.ndarray,
    cell_volume: float,
    atom_positions: np.ndarray,
    atomic_numbers: np.ndarray,
    occupations_core: np.ndarray,
    fft_grid: np.ndarray | None = None,
    use_gpu: bool = True,
) -> float:
    r"""Calculate frozen core energy ERI part
    2\sum_i^{\mathrm{frozen}} h_ii + \sum_{ij}^{\mathrm{frozen}} (2h_{iijj} - h_{ijji})

    Args:
        p (np.ndarray): Array of momentum vectors, shape (#waves, 3)
        c_ip_core (np.ndarray): Array of coefficients describing the active Kohn-Sham orbitals in the plane wave basis, shape (#bands, #waves)
        b (np.ndarray): Array of reciprocal lattice vectors, shape (3, 3)
        mill (np.ndarray): Array of Miller indices, shape (#waves, 3)
        cell_volume (float): Cell volume of the computationen real space cell.
        atom_positions (np.ndarray): 1D array of positions of each atom
        atomic_numbers (np.ndarray): 1D array of atomic number of each atom
        occupations_core (np.ndarray): Array of occupations of the core orbitals.

    Returns:
        np.ndarray: ERIs in reciprocal space
    """
    eri_energy = get_frozen_core_energy_eri(c_ip_core, b, mill, cell_volume, fft_grid, use_gpu=use_gpu)

    # Take only occupied core states into account
    occ = occupations_core.astype(bool)

    iTj_core = calc_matrix_elements.iTj(p, c_ip_core)[occ, :][:, occ]

    # Repulsion between electrons and nuclei
    iUj_core = calc_matrix_elements.iUj(p, c_ip_core, atom_positions, atomic_numbers, cell_volume)[occ, :][:, occ]

    # Frozen core energy
    h_pq_core_trace: float = (iTj_core - iUj_core).trace().real
    frozen_core_energy: float = 2 * h_pq_core_trace + eri_energy

    return frozen_core_energy


def get_frozen_core_energy_pp(
    p: np.ndarray,
    c_ip_core: np.ndarray,
    b: np.ndarray,
    mill: np.ndarray,
    cell_volume: float,
    atom_positions: np.ndarray,
    atomic_numbers: np.ndarray,
    occupations_core: np.ndarray,
    pseudopots: list[Pseudopot],
    fft_grid: np.ndarray | None = None,
    use_gpu: bool = True,
) -> float:
    r"""Calculate frozen core energy ERI part
    2\sum_i^{\mathrm{frozen}} h_ii + \sum_{ij}^{\mathrm{frozen}} (2h_{iijj} - h_{ijji})

    Args:
        p (np.ndarray): Array of momentum vectors, shape (#waves, 3)
        c_ip_core (np.ndarray): Array of coefficients describing the active Kohn-Sham orbitals in the plane wave basis, shape (#bands, #waves)
        b (np.ndarray): Array of reciprocal lattice vectors, shape (3, 3)
        mill (np.ndarray): Array of Miller indices, shape (#waves, 3)
        cell_volume (float): Cell volume of the computationen real space cell.
        atom_positions (np.ndarray): 1D array of positions of each atom
        atomic_numbers (np.ndarray): 1D array of atomic number of each atom
        occupations_core (np.ndarray): Array of occupations of the core orbitals.

    Returns:
        np.ndarray: ERIs in reciprocal space
    """
    eri_energy = get_frozen_core_energy_eri(c_ip_core, b, mill, cell_volume, fft_grid, use_gpu=use_gpu)

    # Take only occupied core states into account
    occ = occupations_core.astype(bool)

    iTj_core = calc_matrix_elements.iTj(p, c_ip_core)[occ, :][:, occ]

    # Repulsion between electrons and nuclei
    pp_core = calc_pseudo_pot.calc_pps(p, c_ip_core, cell_volume, atom_positions, atomic_numbers, pseudopots)[occ, :][:, occ]

    # Frozen core energy
    h_pq_core_trace: float = (iTj_core + pp_core).trace().real
    frozen_core_energy: float = 2 * h_pq_core_trace + eri_energy

    return frozen_core_energy


def get_frozen_core_energy_given_pp(
    p: np.ndarray,
    c_ip_core: np.ndarray,
    b: np.ndarray,
    mill: np.ndarray,
    cell_volume: float,
    pp_core: np.ndarray,
    occupations_core: np.ndarray,
    fft_grid: np.ndarray | None = None,
    use_gpu: bool = True,
) -> float:
    r"""Calculate frozen core energy ERI part
    2\sum_i^{\mathrm{frozen}} h_ii + \sum_{ij}^{\mathrm{frozen}} (2h_{iijj} - h_{ijji})

    Args:
        p (np.ndarray): Array of momentum vectors, shape (#waves, 3)
        c_ip_core (np.ndarray): Array of coefficients describing the active Kohn-Sham orbitals in the plane wave basis, shape (#bands, #waves)
        b (np.ndarray): Array of reciprocal lattice vectors, shape (3, 3)
        mill (np.ndarray): Array of Miller indices, shape (#waves, 3)
        cell_volume (float): Cell volume of the computationen real space cell.
        pp_core (np.ndarray): Pseudopotential matrix for the core orbitals
        occupations_core (np.ndarray): Array of occupations of the core orbitals.

    Returns:
        np.ndarray: ERIs in reciprocal space
    """
    eri_energy = get_frozen_core_energy_eri(c_ip_core, b, mill, cell_volume, fft_grid, use_gpu=use_gpu)

    # Take only occupied core states into account
    occ = occupations_core.astype(bool)

    iTj_core = calc_matrix_elements.iTj(p, c_ip_core)[occ, :][:, occ]

    # Frozen core energy
    h_pq_core_trace: float = (iTj_core + pp_core).trace().real
    frozen_core_energy: float = 2 * h_pq_core_trace + eri_energy

    return frozen_core_energy


def get_frozen_core_pot(
    c_ip_core: np.ndarray,
    c_ip_active: np.ndarray,
    b: np.ndarray,
    mill: np.ndarray,
    fft_grid: np.ndarray | None = None,
    calc_eri_energy: bool = False,
    cell_volume: float | None = None,
    use_gpu: bool = True,
) -> np.ndarray | tuple[np.ndarray, float]:
    r"""Calculate frozen core effective single particle potential V_{tu}=\sum_{i}^{\mathrm{frozen}} (2h_{tuii} - h_{tiiu}).
    The notation is taken from Saad Yalouz et al 2021 Quantum Sci. Technol. 6 024004, Appendix B. Frozen core Hamiltonian,
    i.e. t,u are active orbital indices and i are core orbital indices and chemists index order is used.

    Using physicists' index order this becomes V_{tu}=\sum_{i}^{\mathrm{frozen}} (2h_{tiiu} - h_{tiui}).

    Args:
        c_ip_core (np.ndarray): Array of coefficients describing the core Kohn-Sham orbitals in the plane wave basis, shape (#bands, #waves)
        c_ip_active (np.ndarray): Array of coefficients describing the active Kohn-Sham orbitals in the plane wave basis, shape (#bands, #waves)
        b (np.ndarray): Array of reciprocal lattice vectors, shape (3, 3)
        mill (np.ndarray): Array of Miller indices, shape (#waves, 3)
        calc_eri_energy (bool): Defaults to False.
        cell_volume (float | None): Has to be given if calc_eri_energy is True. Defaults to None.

    Returns:
        np.ndarray | tuple[np.ndarray, float]: ERIs in reciprocal space or
            tuple of (ERIs in reciprocal space, frozen core energy) if calc_eri_energy is True
    """
    if calc_eri_energy:
        assert cell_volume is not None

    # number of core/active DFT bands and number of plane waves
    nbands_core, _nwaves = c_ip_core.shape
    nbands_active, _nwaves = c_ip_active.shape
    dtype = np.complex128

    assert mill.ndim == 2 and mill.shape[1] == 3, f"Array mill should contain 3D Miller indices but has shape {mill.shape}!"
    assert b.ndim == 2 and b.shape[0] == b.shape[1] == 3, f"Array b should contain 3 3D reciprocal lattice vectors but has shape {b.shape}!"

    max_min = mill.max(axis=0) - mill.min(axis=0)
    if fft_grid is not None:
        assert isinstance(fft_grid, np.ndarray), f"fft_grid ({fft_grid}, type {type(fft_grid)}) is not a numpy array!"
        assert fft_grid.shape == (3,), f"fft_grid (shape {fft_grid.shape}) has to have shape (3,)"
        assert np.all(fft_grid >= 0), f"fft_grid {fft_grid} has values smaller than zero!"
        assert np.all(fft_grid >= max_min), f"fft_grid ({fft_grid}) needs to be larger or equalt than wavefunction grid ({max_min})!"

        max_min = fft_grid.copy()

    # Grid needs to have odd edge length values so that the zero momentum is in the center of the grid
    for i, val in enumerate(max_min):
        if val % 2 == 0:
            max_min[i] += 1

    logging.info("FFT grid: %s", max_min)

    logging.info("Init c_ip arrays...")
    # Initialize 3D array for each DFT band with zero
    # We want to define \psi_i(p) as a matrix on the whole momentum grid for each i, instead of a list
    c_ip_core_array = np.zeros(
        (
            nbands_core,
            # Number of grid points for given maximum and minimum Miller index
            *max_min,
        ),
        dtype=dtype,
    )
    c_ip_active_array = np.zeros(
        (
            nbands_active,
            *max_min,
        ),
        dtype=dtype,
    )
    logging.info("Set c_ip arrays...")
    # Set \psi_i(p) on given grid points
    for idx, mill_idx in enumerate(mill):
        x, y, z = mill_idx
        i, j, k = (
            x + max_min[0] // 2,
            y + max_min[1] // 2,
            z + max_min[2] // 2,
        )
        c_ip_core_array[:, i, j, k] = c_ip_core[:, idx]
        c_ip_active_array[:, i, j, k] = c_ip_active[:, idx]

    logging.info("Set p_norm_squared_array...")
    one_over_p_norm_squared_array = np.zeros(max_min)
    one_over_p_norm_squared_array = make_one_over_p_norm_squared_array(one_over_p_norm_squared_array, b, max_min)

    logging.info("Calc pair densities...")
    # Calculate pair density \rho_ij(p) in reciprocal space
    logging.info("Calc pair density core...")
    rho_ij_p_core = pair_density(c_ip_array=c_ip_core_array, use_gpu=use_gpu)
    logging.info("Calc pair density active...")
    rho_tu_p_active = pair_density(c_ip_array=c_ip_active_array, use_gpu=use_gpu)
    logging.info("Calc pair density active-core...")
    rho_ti_p_active_core = pair_density(c_ip_array=c_ip_active_array, c_jp_array=c_ip_core_array, use_gpu=use_gpu)
    # Initialize ERI array
    # TODO: We do not need to calculate all matrix elements due to symmetries
    pot_mat = np.zeros((nbands_active, nbands_active), dtype=dtype)

    logging.info("Calc einsums...")
    # 2 g_tuii - g_tiiu, like in Yalouz et al, but they use chemists order
    # We use physicist order, therefore we use 2 h_tiiu - h_tiui
    # 2 h_tiiu - h_tiui = 2 (4\pi \sum_{p \neq 0} \rho_{tu}(-p)\rho_{ii}(p)/|p|²) - 4\pi \sum_{p \neq 0} \rho_{ti}(-p)\rho_{iu}(p)/|p|²
    #                   = 4\pi \sum_{p \neq 0} [2 \rho_{tu}(-p)\rho_{ii}(p) - \rho_{ti}(-p)\rho_{iu}(p)] /|p|²
    # frozen core potential matrix
    # h_pq_core = 4\pi \sum_{i} \sum_{p \neq 0} [2 \rho_{tu}(-p)\rho_{ii}(p) - \rho_{ti}(-p)\rho_{iu}(p)] /|p|²

    # h_{ijkl}=4\pi \sum_{p \neq 0} \rho_{il}(-p)\rho_{jk}(p)/|p|²
    # h_{tiiu}=4\pi \sum_{p \neq 0} \rho_{tu}(-p)\rho_{ii}(p)/|p|²
    h_tiiu = (
        4
        * np.pi
        * np.einsum(
            "tuxyz, iixyz, xyz -> tiu",  # tiiu, Physicists' order
            np.flip(rho_tu_p_active, (2, 3, 4)) * prod(max_min),
            rho_ij_p_core * prod(max_min),
            one_over_p_norm_squared_array,
            optimize=True,  # larger memory footprint but faster
        )
    )
    # h_{ijkl}=4\pi \sum_{p \neq 0} \rho_{il}(-p)\rho_{jk}(p)/|p|²
    # h_{tiui}=4\pi \sum_{p \neq 0} \rho_{ti}(-p)\rho_{iu}(p)/|p|²
    rho_ti_p_active_core_flip_prod = np.flip(rho_ti_p_active_core, (2, 3, 4)) * prod(max_min)
    h_tiui = (
        4
        * np.pi
        * np.einsum(
            "tixyz, iuxyz, xyz -> tiu",  # tiui, Physicists' order
            rho_ti_p_active_core_flip_prod,
            # \rho_iu(p) = \rho*_ui(-p)
            # swapaxes to bring indices in correct order, np.flip(...) and conj() do p->-p and \rho->\rho*
            #   but not ui -> iu. Regardless of writing \rho_iu(p) or \rho*_ui(-p).
            #   i always runs only over core orbitals while u always runs over active orbitals
            rho_ti_p_active_core_flip_prod.conj().swapaxes(0, 1),  # \rho_iu(p) = \rho*_ui(-p)
            one_over_p_norm_squared_array,
            optimize=True,  # larger memory footprint but faster
        )
    )

    pot_mat = 2 * h_tiiu.sum(1) - h_tiui.sum(1)

    if calc_eri_energy:
        eri_energy = get_frozen_core_energy_eri(
            c_ip_core=c_ip_core,
            b=b,
            mill=mill,
            cell_volume=cell_volume,
            fft_grid=max_min,
            rho_ij_p=rho_ij_p_core,
            one_over_p_norm_squared_array=one_over_p_norm_squared_array,
            use_gpu=use_gpu,
        )
        return pot_mat, eri_energy

    return pot_mat


def get_frozen_core_pot_and_energy_given_pp(
    c_ip_core: np.ndarray,
    c_ip_active: np.ndarray,
    b: np.ndarray,
    mill: np.ndarray,
    p: np.ndarray,
    cell_volume: float,
    pp_core: np.ndarray,
    occupations_core: np.ndarray,
    fft_grid: np.ndarray | None = None,
    use_gpu: bool = True,
) -> tuple[np.ndarray, float]:
    pot_mat, eri_energy = get_frozen_core_pot(
        c_ip_core=c_ip_core,
        c_ip_active=c_ip_active,
        b=b,
        mill=mill,
        fft_grid=fft_grid,
        calc_eri_energy=True,
        cell_volume=cell_volume,
        use_gpu=use_gpu,
    )

    # Take only occupied core states into account
    occ = occupations_core.astype(bool)

    iTj_core = calc_matrix_elements.iTj(p, c_ip_core)[occ, :][:, occ]

    # Frozen core energy
    h_pq_core_trace: float = (iTj_core + pp_core).trace().real
    frozen_core_energy: float = 2 * h_pq_core_trace + eri_energy

    return pot_mat / cell_volume, frozen_core_energy



import os
import copy
import numpy as np
import dopyqo
from dopyqo.colors import *
from dopyqo.units import Unit
from dopyqo import DopyqoConfig, MatrixElements, print_block
from dopyqo.scripts.main import run as run_dopyqo

if __name__ == "__main__":
    # From QE output
    base_folder = os.path.join("qe_files", "Be")
    prefix = "Be"
    xml_file = os.path.join(base_folder, f"{prefix}.save", "data-file-schema.xml")
    wfc_file = os.path.join(base_folder, f"{prefix}.save", "wfc1.dat")
    wfc_obj = dopyqo.Wfc.from_file(wfc_file, xml_file)
    wfc_obj.set_atom_positions(wfc_obj.atom_positions_hartree + 0.01, unit=Unit.HARTREE)
    wfc_obj.set_lattice_vectors(wfc_obj.a + 0.5, unit=Unit.HARTREE)
    wfc_obj.nbnd = 20
    input_file = os.path.join("qe_files", "Be", "Be_from_dopyqo.scf.in")
    wfc_obj.to_qe_input(input_file, prefix="Be")

    # # Or from dopyqo calculation
    # config = DopyqoConfig(
    #     base_folder="Be",
    #     prefix="Be",
    #     active_electrons=2,
    #     active_orbitals=2,
    #     run_fci=False,
    #     run_vqe=False,
    #     n_threads=10,
    # )
    # energy_dict, wfc_obj, h_ks, mats = run_dopyqo(config, return_h=True, return_wfc=True, return_mats=True)

    # wfc_obj.set_atom_positions(wfc_obj.atom_positions_hartree + 0.01, unit=Unit.HARTREE)
    # wfc_obj.to_qe_input(os.path.join("Be_simple", "Be_simple.scf.in"), prefix="Be_simple")

    # # Run QE SCF calculation
    # dopyqo.runQE(input_file)



import os
import sys
import numpy as np
from dopyqo.colors import *
from dopyqo import DopyqoConfig


def read_u_mat(filename: str) -> dict[tuple[float, float, float], np.ndarray]:
    # Initialize a list to store the data for each k-point
    kpt_to_u = {}

    # Open file for reading
    with open(filename, "r", encoding="utf-8") as f:
        # Read the first line containing the date and time of creation of the file
        _creation_time = f.readline().strip()

        # Read the second line containing the number of k-points (num_kpts) and two times the number of Wannier functions (num_wann)
        num_kpts, num_wann, _num_wann = map(int, f.readline().split())

        # Skip the third line which is empty
        f.readline()

        # Read the data for each k-point
        for _ in range(num_kpts):
            # Read the k-point coordinates (in fractional coordinates of the reciprocal lattice vectors)
            kpt = tuple(map(float, f.readline().split()))

            # Initialize a matrix to store the U(k) matrix elements
            u_mat = np.zeros(shape=(num_wann, num_wann), dtype=np.complex128)

            # Read the matrix elements (real and imaginary parts) of U(k) in column-major order,
            # i.e. cycling over rows first and then columns
            # Note: \psi^{(k)}_{n, wannier} = U^{(k)}_{mn} \psi^{(k)}_{m, KS}
            for col in range(num_wann):
                for row in range(num_wann):
                    real, imag = map(float, f.readline().split())
                    u_mat[row, col] = complex(real, imag)

            # Append the data for this k-point to the list
            # kpt_to_u[kpt] = u_mat
            # NOTE: We have to transform with U*^T instead of with U to replicate
            #       the XSF data of the real-space Wannier orbitals (test for real-wavefunctions at gamma-point).
            kpt_to_u[kpt] = u_mat.T.conj()

            # Skip empty line between k-points
            f.readline()

    return kpt_to_u


# def get_u_mat(config: DopyqoConfig, kpoint: np.ndarray, nbnd: int, orbital_indices_active: list[int]):
#     # Read Wannier90 transform matrix
#     transform_matrix = read_u_mat(config.wannier_umat)[tuple(kpoint)]
#     if not check_unitarity_u(transform_matrix):
#         print(f"{RED}Wannier error: Transformation matrix is not unitary{RESET_COLOR}")
#         sys.exit(1)
#     if config.active_orbitals != transform_matrix.shape[0]:
#         print(
#             f"{RED}Wannier error: Number of active orbitals ({config.active_orbitals}) "
#             + f"does not match number of Wannier orbitals ({transform_matrix.shape[0]})!{RESET_COLOR}"
#         )
#         sys.exit(1)
#     ######################### READING WANNIER INPUT FILE #########################
#     if config.wannier_input_file is not None:
#         results = {}
#         keys = ["num_wann", "exclude_bands"]

#         with open(config.wannier_input_file, "r", encoding="utf-8") as f:
#             for line in f:
#                 line = line.strip()

#                 for key in keys:
#                     if line.startswith(key):
#                         if "=" in line:
#                             _, sep, val = line.partition("=")
#                         elif ":" in line:
#                             _, sep, val = line.partition(":")
#                         else:  # whitespace split
#                             parts = line.split(None, 1)
#                             val = parts[1] if len(parts) == 2 else None
#                         results[key] = val.strip()
#                         break

#         _num_wann = int(results["num_wann"])
#         exclude_bands = []
#         if "exclude_bands" in results:
#             for x in results["exclude_bands"].split(","):
#                 if "-" not in x:
#                     exclude_bands.append(int(x))
#                 else:
#                     start_stop = [int(val) for val in x.split("-")]
#                     start = start_stop[0]
#                     stop = start_stop[1] + 1
#                     exclude_bands.extend(list(range(start, stop)))
#             exclude_bands = [x - 1 for x in exclude_bands]  # Wannier90 1-indexing to python 0-indexing
#         orbital_indices_active_wannier = list(set(exclude_bands) ^ set(range(nbnd)))
#         if orbital_indices_active != orbital_indices_active_wannier:
#             print(
#                 f"{RED}Wannier error: Orbitals in the active space ({orbital_indices_active}) are not the "
#                 + f"same orbitals transformed with Wannier90 ({orbital_indices_active_wannier})!{RESET_COLOR}"
#             )
#             sys.exit(1)


def read_hr_dat(filename: str) -> dict[tuple[float, float, float], np.ndarray]:
    # Open file for reading
    with open(filename, "r", encoding="utf-8") as f:
        # Read the first line containing the date and time of creation of the file
        _creation_time = f.readline().strip()

        num_wann = int(f.readline().strip())
        nrpts = int(f.readline().strip())

        degeneracies = []
        for _ in range(nrpts // 15 + 1):
            degeneracies.append(f.readline().split())

        hr_mat_ev = np.zeros(shape=(num_wann, num_wann, nrpts), dtype=np.complex128)

        for i in range(num_wann**2 * nrpts):
            _r_1, _r_2, _r_3, m, n, real, imag = map(float, f.readline().split())
            hr_mat_ev[int(m) - 1, int(n) - 1, i // num_wann**2] = complex(real, imag)

    ev_to_hartree = 0.036749308136649
    hr_mat_hartree = hr_mat_ev * ev_to_hartree
    return hr_mat_hartree


# From https://github.com/hungpham2017/mcu/blob/master/mcu/wannier90/utils.py#L53
# referenced here: https://lists.quantum-espresso.org/pipermail/wannier/2020-March/001726.html
def read_U_matrix(filename):
    """Read seedname_u.mat file"""

    with open(filename, "r", encoding="utf-8") as file:
        data = file.read().split("\n")
        nkpts, nwann, nband = np.int64(data[1].split())
        temp = data[2:-1]
        block_length = nband * nwann + 2
        kpts = []
        U_kpts = []
        for kpt_th in range(nkpts):
            Uk = temp[(kpt_th * block_length) : (kpt_th * block_length + block_length)]
            kpts.append(np.float64(Uk[1].split()))
            U = np.asarray([np.float64(line.split()[0]) + 1j * np.float64(line.split()[1]) for line in Uk[2:]])
            U_kpts.append(U.reshape(nwann, nband).T)

        kpts = np.float64(kpts)
        U_kpts = np.asarray(U_kpts)

    return kpts, U_kpts


def check_unitarity_u(u_mat: np.ndarray) -> bool:
    """Check unitarity of U matrix
    u_mat: U matrix at different k-point with shape (nwann, nwann, nkpts)
    """
    assert u_mat.ndim == 2
    assert u_mat.shape[0] == u_mat.shape[1]
    nwann = u_mat.shape[0]

    return np.allclose(u_mat @ u_mat.T.conj(), np.eye(nwann))


if __name__ == "__main__":
    filename_umat_mg = os.path.join("wannier90_files", "magnesium_u.mat")

    kpt_to_u_mg = read_u_mat(filename=filename_umat_mg)

    filename_hr_mg = os.path.join("wannier90_files", "calc", "magnesium_hr.dat")

    hr_mat_mg = read_hr_dat(filename=filename_hr_mg)
    # KS eigenvalues and eigenvectors in the Wannier basis
    eigvals, eigvecs = np.linalg.eig(hr_mat_mg[:, :, 0])

    # Check unitarity
    u_mat_mg = np.stack(list(kpt_to_u_mg.values()), axis=2)
    print(f"Unitary: {check_unitarity_u(u_mat_mg)}")
